# Boltlight - a LN node wrapper
#
# Copyright (C) 2021-2022 boltlight contributors
# Copyright (C) 2018 inbitcoin s.r.l.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For a full list of contributors, please see the AUTHORS.md file.
"""Implementation of boltlight.proto defined methods for c-lightning."""

from ast import literal_eval
from concurrent.futures import ThreadPoolExecutor
from concurrent.futures import TimeoutError as TimeoutFutError
from contextlib import suppress
from datetime import datetime
from logging import CRITICAL, getLogger
from os import path

from pyln.client import LightningRpc
from pyln.client import RpcError as ClightningRpcError

from . import boltlight_pb2 as pb
from . import settings as sett
from .errors import Err
from .utils.bitcoin import Enforcer as Enf
from .utils.bitcoin import (
    get_channel_balances, has_amount_encoded, split_node_uri)
from .utils.misc import get_path, handle_thread, set_defaults
from .utils.network import FakeContext, check_req_params, get_thread_timeout

LOGGER = getLogger(__name__)

ERRORS = {
    "'fundchannel_start' failed": {
        'fun': 'openchannel_failed'
    },
    'Bad bech32 string': {
        'fun': 'invalid',
        'params': 'payment_request'
    },
    'Cannot afford transaction': {
        'fun': 'insufficient_funds'
    },
    'Channel ID not found': {
        'fun': 'invalid',
        'params': 'channel_id'
    },
    'Connection refused': {
        'fun': 'node_error'
    },
    'Could not find a route': {
        'fun': 'route_not_found'
    },
    'does not match description': {
        'fun': 'incorrect_description'
    },
    'Error broadcasting transaction': {
        'fun': 'payonchain_failed'
    },
    'Exchanging init messages: Operation now in progress': {
        'fun': 'connect_failed'
    },
    'Fallback address does not match our network': {
        'fun': 'invalid',
        'params': 'fallback_addr'
    },
    'Fallback address not valid': {
        'fun': 'invalid',
        'params': 'fallback_addr'
    },
    'Given id is not a channel ID or short channel ID': {
        'fun': 'invalid',
        'params': 'channel_id'
    },
    # this error happens when giving a short fallback address (e.g. "sd")
    'Incorrect \'id\' in response': {
        'fun': 'invalid',
        'params': 'fallback_addr'
    },
    'Invoice expired': {
        'fun': 'invoice_expired',
    },
    'msatoshi parameter required': {
        'fun': 'amount_required'
    },
    'no description to check': {
        'fun': 'missing_parameter',
        'params': 'description'
    },
    'Parsing accept_channel': {
        'fun': 'openchannel_failed'
    },
    'payment_hash: should be a 32 byte hex value': {
        'fun': 'invalid',
        'params': 'payment_hash'
    },
    'Peer already': {
        'fun': 'openchannel_failed'
    },
    'Still syncing with bitcoin network': {
        'fun': 'openchannel_failed'
    },
    'They sent error': {
        'fun': 'openchannel_failed'
    },
    'Unknown peer': {
        'fun': 'connect_failed'
    },
}


def get_node_version():
    """Get node's version."""
    rpc_cl = ClightningRPC()
    with suppress(RuntimeError):
        cl_res, is_err = rpc_cl.getinfo(FakeContext())
        if not is_err:
            return cl_res.get('version', '')
    return ''


def get_settings(config, sec):
    """Get c-lightning settings."""
    cl_values = ['CL_RPC']
    set_defaults(config, cl_values)
    cl_rpc_dir = get_path(config.get(sec, 'CL_RPC_DIR'))
    cl_rpc = config.get(sec, 'CL_RPC')
    cl_rpc_path = path.join(cl_rpc_dir, cl_rpc)
    if not path.exists(cl_rpc_path):
        raise RuntimeError(f'Missing {cl_rpc} file')
    sett.RPC_URL = cl_rpc_path


def unlock_node(_ctx, _password, _session=None):
    """Return a successful response (no locking system for c-lightning)."""
    return pb.UnlockNodeResponse()


def update_settings(_dummy):
    """Update c-lightning specific settings."""


def BalanceOffChain(_req, ctx):
    """Return the off-chain balance available across all channels."""
    channels = ListChannels(pb.ListChannelsRequest(), ctx).channels
    return get_channel_balances(channels)


def BalanceOnChain(_req, ctx):
    """Return the on-chain balance in satoshi of the running LN node."""
    rpc_cl = ClightningRPC()
    cl_res, is_err = rpc_cl.listfunds(ctx)
    if is_err:
        _handle_error(ctx, cl_res)
    tot_funds = conf_funds = 0.0
    for output in cl_res.get('outputs', []):
        output_value = output.get('value', 0)
        tot_funds += output_value
        if output.get('status') == 'confirmed':
            conf_funds += output_value
    return pb.BalanceOnChainResponse(total_sat=int(tot_funds),
                                     confirmed_sat=int(conf_funds))


def CheckInvoice(req, ctx):
    """Check if a LN invoice has been paid."""
    rpc_cl = ClightningRPC()
    check_req_params(ctx, req, 'payment_hash')
    invoice = None
    cl_req = {'payment_hash': req.payment_hash}
    cl_res, is_err = rpc_cl.listinvoices(ctx, cl_req)
    if is_err:
        _handle_error(ctx, cl_res)
    for inv in cl_res.get('invoices', []):
        if inv.get('payment_hash') == req.payment_hash:
            invoice = inv
            break
    if not invoice:
        Err().invoice_not_found(ctx)
    return pb.CheckInvoiceResponse(state=_get_invoice_state(invoice))


def CloseChannel(req, ctx):
    """Try to close a LN channel."""
    check_req_params(ctx, req, 'channel_id')
    cl_req = {'peer_id': req.channel_id}
    if req.force:
        # setting a 1 second timeout to force an immediate unilateral close
        cl_req['unilateraltimeout'] = 1
    with ThreadPoolExecutor(max_workers=1) as executor:
        future = executor.submit(_close_channel, cl_req)
        try:
            cl_res = future.result(timeout=get_thread_timeout(ctx))
            if cl_res:
                return pb.CloseChannelResponse(closing_txid=cl_res.get('txid'))
        except RuntimeError as cl_err:
            try:
                error = literal_eval(str(cl_err))
                _handle_error(ctx, error)
            except (SyntaxError, ValueError):
                Err().report_error(ctx, str(cl_err))
        except TimeoutFutError:
            executor.shutdown(wait=False)
    return pb.CloseChannelResponse()


def CreateInvoice(req, ctx):
    """Create a LN invoice (bolt 11)."""
    rpc_cl = ClightningRPC()
    cl_req = {}
    if req.min_final_cltv_expiry:
        cl_req['cltv'] = req.min_final_cltv_expiry
    if req.amount_msat and Enf.check_value(
            ctx, req.amount_msat, enforce=Enf.LN_PAYREQ):
        cl_req['msatoshi'] = req.amount_msat
    else:
        cl_req['msatoshi'] = 'any'
    description = ''
    if req.description:
        description = req.description
    cl_req['description'] = description
    label = _create_label()
    cl_req['label'] = label
    if req.expiry:
        cl_req['expiry'] = req.expiry
    else:
        cl_req['expiry'] = sett.EXPIRY_TIME
    if req.fallback_addr:
        cl_req['fallbacks'] = [req.fallback_addr]
    cl_res, is_err = rpc_cl.invoice(ctx, cl_req)
    if is_err:
        _handle_error(ctx, cl_res)
    return pb.CreateInvoiceResponse(payment_hash=cl_res.get('payment_hash'),
                                    payment_request=cl_res.get('bolt11'),
                                    expires_at=cl_res.get('expires_at'))


def DecodeInvoice(req, ctx):
    """Return information of an invoice from its payment request (bolt 11)."""
    rpc_cl = ClightningRPC()
    check_req_params(ctx, req, 'payment_request')
    cl_req = {'bolt11': req.payment_request}
    if req.description:
        cl_req['description'] = req.description
    cl_res, is_err = rpc_cl.decodepay(ctx, cl_req)
    if is_err:
        _handle_error(ctx, cl_res)
    res = pb.DecodeInvoiceResponse(
        amount_msat=int(cl_res.get('msatoshi', 0)),
        timestamp=cl_res.get('created_at'),
        payment_hash=cl_res.get('payment_hash'),
        description=cl_res.get('description'),
        destination_pubkey=cl_res.get('payee'),
        description_hash=cl_res.get('description_hash'),
        expiry=cl_res.get('expiry'),
        min_final_cltv_expiry=cl_res.get('min_final_cltv_expiry'))
    fallbacks = cl_res.get('fallbacks', [])
    if len(fallbacks):
        res.fallback_addr = fallbacks[0].get('addr', '')
    for cl_route in cl_res.get('routes', []):
        _add_route_hint(res, cl_route)
    return res


def GetNodeInfo(_req, ctx):
    """Return info about the running LN node."""
    rpc_cl = ClightningRPC()
    cl_res, is_err = rpc_cl.getinfo(ctx)
    if is_err:
        _handle_error(ctx, cl_res)
    color = cl_res.get('color')
    res = pb.GetNodeInfoResponse(identity_pubkey=cl_res.get('id'),
                                 alias=cl_res.get('alias'),
                                 block_height=int(cl_res.get('blockheight',
                                                             0)),
                                 color=f'#{color}' if color else '')
    if res.identity_pubkey:
        addresses = cl_res.get('address', [])
        if len(addresses):
            address = addresses[0]
            addr = address.get('address')
            port = address.get('port')
            if addr and port:
                res.node_uri = f'{res.identity_pubkey}@{addr}:{port}'
    network = cl_res.get('network')
    if network == 'bitcoin':
        res.network = pb.Network.MAINNET
    elif network == 'testnet':
        res.network = pb.Network.TESTNET
    elif network == 'regtest':
        res.network = pb.Network.REGTEST
    return res


def ListChannels(req, ctx):
    """Return a list of channels of the running LN node."""
    rpc_cl = ClightningRPC()
    cl_res, is_err = rpc_cl.listpeers(ctx)
    if is_err:
        _handle_error(ctx, cl_res)
    res = pb.ListChannelsResponse()
    for cl_peer in cl_res.get('peers', []):
        for cl_chan in cl_peer.get('channels', []):
            _add_channel(res, cl_peer, cl_chan, req.active_only)
    return res


def ListPayments(_req, ctx):
    """Return a list of lightning invoices paid by the running LN node."""
    rpc_cl = ClightningRPC()
    cl_res, is_err = rpc_cl.listsendpays(ctx)
    if is_err:
        _handle_error(ctx, cl_res)
    res = pb.ListPaymentsResponse()
    for cl_payment in cl_res.get('payments', []):
        _add_payment(res, cl_payment)
    return res


def ListPeers(_req, ctx):
    """Return a list of peers connected to the running LN node."""
    rpc_cl = ClightningRPC()
    cl_res, is_err = rpc_cl.listpeers(ctx)
    if is_err:
        _handle_error(ctx, cl_res)
    res = pb.ListPeersResponse()
    for peer in cl_res.get('peers', []):
        # Filtering disconnected peers
        if peer.get('connected') is not True:
            continue
        grpc_peer = res.peers.add(pubkey=peer.get('id'))
        if grpc_peer.pubkey:
            cl_req = {'node_id': grpc_peer.pubkey}
            cl_res, is_err = rpc_cl.listnodes(ctx, cl_req)
            nodes = cl_res.get('nodes', [])
            if len(nodes):
                node = nodes[0]
                grpc_peer.alias = node.get('alias', '')
                color = node.get('color')
                grpc_peer.color = f'#{color}' if color else ''
        addresses = peer.get('netaddr', [])
        if len(addresses):
            grpc_peer.address = addresses[0]
    return res


def NewAddress(req, ctx):
    """Create a new bitcoin address under control of the running LN node."""
    rpc_cl = ClightningRPC()
    cl_req = {}
    if req.addr_type == pb.Address.P2WPKH:
        cl_req['addresstype'] = 'bech32'
    elif req.addr_type == pb.Address.NP2WPKH:
        cl_req['addresstype'] = 'p2sh-segwit'
    cl_res, is_err = rpc_cl.newaddr(ctx, cl_req)
    if is_err:
        _handle_error(ctx, cl_res)
    res = pb.NewAddressResponse(address=cl_res.get('bech32'))
    if not res.address:
        res.address = cl_res.get('p2sh-segwit', '')
    return res


def OpenChannel(req, ctx):
    """Try to connect and open a channel with a peer."""
    check_req_params(ctx, req, 'node_uri', 'funding_sat')
    pubkey, _ = split_node_uri(ctx, req.node_uri)
    rpc_cl = ClightningRPC()
    cl_req = {'peer_id': req.node_uri}
    cl_res, is_err = rpc_cl.connect(ctx, cl_req)
    if is_err:
        Err().connect_failed(ctx)
    Enf.check_value(ctx, req.funding_sat, Enf.FUNDING_SATOSHIS)
    cl_req = {'node_id': pubkey, 'amount': req.funding_sat}
    if req.private:
        cl_req['announce'] = False
    if req.push_msat and Enf.check_value(ctx, req.push_msat, Enf.PUSH_MSAT):
        if req.push_msat >= 1000 * req.funding_sat:
            Err().value_too_high(ctx, req.push_msat)
        cl_req['push_msat'] = req.push_msat
    cl_res, is_err = rpc_cl.fundchannel(ctx, cl_req)
    if is_err:
        _handle_error(ctx, cl_res)
    return pb.OpenChannelResponse(funding_txid=cl_res.get('txid'))


def PayInvoice(req, ctx):
    """Try to pay a LN invoice from its payment request (bolt 11).

    An amount can be specified if the invoice doesn't already have it included.
    If a description hash is included in the invoice, its preimage must be
    included in the request.
    """
    check_req_params(ctx, req, 'payment_request')
    rpc_cl = ClightningRPC()
    cl_req = {'bolt11': req.payment_request}
    amount_encoded = has_amount_encoded(req.payment_request)
    if amount_encoded and req.amount_msat:
        Err().unsettable(ctx, 'amount_msat')
    elif req.amount_msat and not amount_encoded:
        Enf.check_value(ctx, req.amount_msat, enforce=Enf.LN_TX)
        cl_req['msatoshi'] = req.amount_msat
    elif not amount_encoded:
        check_req_params(ctx, req, 'amount_msat')
    if req.description:
        Err().unimplemented_parameter(ctx, 'description')
    if req.cltv_expiry_delta:
        Enf.check_value(ctx,
                        req.cltv_expiry_delta,
                        enforce=Enf.CLTV_EXPIRY_DELTA)
        cl_req['maxdelay'] = req.cltv_expiry_delta
    cl_res, is_err = rpc_cl.pay(ctx, cl_req)
    if is_err:
        _handle_error(ctx, cl_res)
    return pb.PayInvoiceResponse(
        payment_preimage=cl_res.get('payment_preimage'))


def PayOnChain(req, ctx):
    """Try to pay a bitcoin address."""
    rpc_cl = ClightningRPC()
    check_req_params(ctx, req, 'address', 'amount_sat')
    Enf.check_value(ctx, req.amount_sat, enforce=Enf.OC_TX)
    cl_req = {'destination': req.address, 'satoshi': req.amount_sat}
    if req.fee_sat_byte:
        Enf.check_value(ctx, req.fee_sat_byte, enforce=Enf.OC_FEE)
        cl_req['feerate'] = f'{req.fee_sat_byte * 1000}perkb'
    cl_res, is_err = rpc_cl.withdraw(ctx, cl_req)
    if is_err:
        _handle_error(ctx, cl_res)
    return pb.PayOnChainResponse(txid=cl_res.get('txid'))


def UnlockNode(_req, _ctx):
    """Try to unlock node."""
    return unlock_node(None, None)


def _add_channel(res, cl_peer, cl_chan, active_only):
    """Add a channel to a ListChannelsResponse."""
    state = _get_channel_state(cl_chan)
    active = state == pb.Channel.OPEN and cl_peer.get('connected')
    if state < 0 or (active_only and not active):
        return
    grpc_chan = res.channels.add(
        active=active,
        capacity_msat=int(cl_chan.get('msatoshi_total', 0)),
        channel_id=cl_chan.get('channel_id'),
        funding_txid=cl_chan.get('funding_txid'),
        local_balance_msat=int(cl_chan.get('msatoshi_to_us', 0)),
        local_reserve_sat=cl_chan.get('our_channel_reserve_satoshis'),
        private=cl_chan.get('private'),
        remote_pubkey=cl_peer.get('id'),
        remote_reserve_sat=cl_chan.get('their_channel_reserve_satoshis'),
        short_channel_id=cl_chan.get('short_channel_id'),
        state=state,
        to_self_delay=int(cl_chan.get('our_to_self_delay', 0)))
    grpc_chan.remote_balance_msat = \
        grpc_chan.capacity_msat - grpc_chan.local_balance_msat


def _add_payment(res, cl_payment):
    """Add a payment to a ListPaymentsResponse."""
    if cl_payment.get('status') == 'failed':
        return
    grpc_payment = res.payments.add(
        payment_hash=cl_payment.get('payment_hash'),
        amount_msat=int(cl_payment.get('msatoshi_sent', 0)),
        timestamp=cl_payment.get('created_at'),
        payment_preimage=cl_payment.get('payment_preimage'))
    grpc_payment.fee_msat = max(
        0, grpc_payment.amount_msat - int(cl_payment.get('msatoshi', 0)))


def _add_route_hint(res, cl_route):
    """Add a route hint and its hop hints to a DecodeInvoiceResponse."""
    grpc_route = res.route_hints.add()
    for cl_hop in cl_route:
        grpc_route.hop_hints.add(
            pubkey=cl_hop.get('pubkey'),
            short_channel_id=cl_hop.get('short_channel_id'),
            fee_base_msat=cl_hop.get('fee_base_msat'),
            fee_proportional_millionths=cl_hop.get(
                'fee_proportional_millionths'),
            cltv_expiry_delta=cl_hop.get('cltv_expiry_delta'))


@handle_thread
def _close_channel(cl_req):
    """Close a LN channel and return a closing TXID or raise an exception."""
    rpc_cl = ClightningRPC()
    cl_res = error = None
    try:
        cl_res, is_err = rpc_cl.close(FakeContext(), cl_req)
        if is_err:
            error = cl_res
        else:
            LOGGER.debug('[ASYNC] CloseChannel terminated with response: %s',
                         cl_res)
    except RuntimeError as err:
        error = str(err)
    if error:
        LOGGER.debug('[ASYNC] CloseChannel terminated with error: %s', error)
        raise RuntimeError(error)
    return cl_res


def _create_label():
    """Create a label using microseconds (c-lightning specific)."""
    return str(int(datetime.now().timestamp() * 1e6))


def _get_channel_state(cl_chan):  # pylint: disable=too-many-return-statements
    """Return the channel state."""
    cl_state = cl_chan.get('state')
    if not cl_state or cl_state in ('CLOSED', ):
        return -1
    cl_state_details = cl_chan.get('status', [])
    for detail in cl_state_details:
        if 'ONCHAIN:All outputs resolved:' in detail:
            return -1
    if cl_state in ('CHANNELD_AWAITING_LOCKIN', ):
        return pb.Channel.PENDING_OPEN
    if cl_state in ('CHANNELD_NORMAL', ):
        return pb.Channel.OPEN
    for detail in cl_state_details:
        if 'ONCHAIN:Tracking mutual close transaction' in detail:
            return pb.Channel.PENDING_MUTUAL_CLOSE
        if 'ONCHAIN:Tracking our own unilateral close' in detail or \
                'ONCHAIN:2 outputs unresolved:' in detail or \
                'ONCHAIN:1 outputs unresolved:' in detail or \
                'ONCHAIN:Tracking their unilateral close' in detail:
            return pb.Channel.PENDING_FORCE_CLOSE
    if cl_state in ('CHANNELD_SHUTTING_DOWN', 'CLOSINGD_SIGEXCHANGE',
                    'CLOSINGD_COMPLETE'):
        return pb.Channel.PENDING_MUTUAL_CLOSE
    if cl_state in ('ONCHAIN', 'AWAITING_UNILATERAL', 'FUNDING_SPEND_SEEN'):
        return pb.Channel.PENDING_FORCE_CLOSE
    return pb.Channel.UNKNOWN


def _get_invoice_state(cl_invoice):
    """Return the invoice state."""
    status = cl_invoice.get('status')
    if status == 'paid':
        return pb.Invoice.PAID
    if status == 'unpaid':
        return pb.Invoice.PENDING
    if status == 'expired':
        return pb.Invoice.EXPIRED
    return pb.Invoice.UNKNOWN


def _handle_error(ctx, cl_res):
    """Report errors of a c-lightning RPC response.

    This is always terminating: raises a grpc.RpcError to the API caller.
    """
    Err().report_error(ctx, cl_res)


class ClightningRPC():  # pylint: disable=too-few-public-methods
    """Create and mantain an RPC session with c-lightning."""
    def __init__(self):
        logger = getLogger(self.__class__.__name__)
        logger.setLevel(CRITICAL)
        self._session = LightningRpc(sett.RPC_URL, logger=logger)

    def __getattr__(self, name):
        def call_adapter(ctx, params=None):
            if not params:
                params = {}
            LOGGER.debug("RPC req: '%s' '%s'", name, params)
            try:
                res = getattr(self._session, name)(**params)
                LOGGER.debug('RPC res: %s', res)
                return res, False
            except ClightningRpcError as err:
                err_msg = err.error.get('message')
                LOGGER.debug('RPC err: %s', err_msg)
                return err_msg, True
            except OSError as err:
                return Err().node_error(ctx, str(err))

        return call_adapter



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: boltlight/boltlight.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x19\x62oltlight/boltlight.proto\x12\tboltlight\"6\n\rUnlockRequest\x12\x10\n\x08password\x18\x01 \x01(\t\x12\x13\n\x0bunlock_node\x18\x02 \x01(\x08\"\'\n\x0eUnlockResponse\x12\x15\n\rnode_unlocked\x18\x01 \x01(\x08\"\x10\n\x0eGetInfoRequest\"U\n\x0fGetInfoResponse\x12\x0f\n\x07version\x18\x01 \x01(\t\x12\x1b\n\x13node_implementation\x18\x02 \x01(\t\x12\x14\n\x0cnode_version\x18\x03 \x01(\t\"\r\n\x0bLockRequest\"\x0e\n\x0cLockResponse\"\x18\n\x16\x42\x61lanceOffChainRequest\"w\n\x17\x42\x61lanceOffChainResponse\x12\x14\n\x0cout_tot_msat\x18\x01 \x01(\x04\x12\x18\n\x10out_tot_now_msat\x18\x02 \x01(\x04\x12\x13\n\x0bin_tot_msat\x18\x04 \x01(\x04\x12\x17\n\x0fin_tot_now_msat\x18\x05 \x01(\x04\"+\n\x13\x43heckInvoiceRequest\x12\x14\n\x0cpayment_hash\x18\x01 \x01(\t\"?\n\x14\x43heckInvoiceResponse\x12\'\n\x05state\x18\x01 \x01(\x0e\x32\x18.boltlight.Invoice.State\"8\n\x13\x43loseChannelRequest\x12\x12\n\nchannel_id\x18\x01 \x01(\t\x12\r\n\x05\x66orce\x18\x02 \x01(\x08\",\n\x14\x43loseChannelResponse\x12\x14\n\x0c\x63losing_txid\x18\x01 \x01(\t\"\x86\x01\n\x14\x43reateInvoiceRequest\x12\x13\n\x0b\x61mount_msat\x18\x01 \x01(\x04\x12\x13\n\x0b\x64\x65scription\x18\x02 \x01(\t\x12\x0e\n\x06\x65xpiry\x18\x03 \x01(\r\x12\x1d\n\x15min_final_cltv_expiry\x18\x04 \x01(\r\x12\x15\n\rfallback_addr\x18\x05 \x01(\t\"Z\n\x15\x43reateInvoiceResponse\x12\x17\n\x0fpayment_request\x18\x01 \x01(\t\x12\x14\n\x0cpayment_hash\x18\x02 \x01(\t\x12\x12\n\nexpires_at\x18\x03 \x01(\x04\"\x14\n\x12GetNodeInfoRequest\"D\n\x14\x44\x65\x63odeInvoiceRequest\x12\x17\n\x0fpayment_request\x18\x01 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x02 \x01(\t\"\x91\x02\n\x15\x44\x65\x63odeInvoiceResponse\x12\x13\n\x0b\x61mount_msat\x18\x01 \x01(\x04\x12\x14\n\x0cpayment_hash\x18\x02 \x01(\t\x12\x1a\n\x12\x64\x65stination_pubkey\x18\x03 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x04 \x01(\t\x12\x18\n\x10\x64\x65scription_hash\x18\x05 \x01(\t\x12\x11\n\ttimestamp\x18\x06 \x01(\x04\x12\x0e\n\x06\x65xpiry\x18\x07 \x01(\r\x12\x1d\n\x15min_final_cltv_expiry\x18\x08 \x01(\r\x12\x15\n\rfallback_addr\x18\t \x01(\t\x12)\n\x0broute_hints\x18\n \x03(\x0b\x32\x14.boltlight.RouteHint\"2\n\tRouteHint\x12%\n\thop_hints\x18\x01 \x03(\x0b\x32\x12.boltlight.HopHint\"\x8a\x01\n\x07HopHint\x12\x0e\n\x06pubkey\x18\x01 \x01(\t\x12\x18\n\x10short_channel_id\x18\x02 \x01(\t\x12\x15\n\rfee_base_msat\x18\x03 \x01(\r\x12#\n\x1b\x66\x65\x65_proportional_millionths\x18\x04 \x01(\r\x12\x19\n\x11\x63ltv_expiry_delta\x18\x05 \x01(\r\"E\n\x07Network\":\n\x04Name\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x0b\n\x07MAINNET\x10\x01\x12\x0b\n\x07TESTNET\x10\x02\x12\x0b\n\x07REGTEST\x10\x03\"\x9e\x01\n\x13GetNodeInfoResponse\x12\x10\n\x08node_uri\x18\x01 \x01(\t\x12\x17\n\x0fidentity_pubkey\x18\x02 \x01(\t\x12\r\n\x05\x61lias\x18\x03 \x01(\t\x12\r\n\x05\x63olor\x18\x04 \x01(\t\x12\x14\n\x0c\x62lock_height\x18\x05 \x01(\r\x12(\n\x07network\x18\x06 \x01(\x0e\x32\x17.boltlight.Network.Name\"*\n\x13ListChannelsRequest\x12\x13\n\x0b\x61\x63tive_only\x18\x01 \x01(\x08\"<\n\x14ListChannelsResponse\x12$\n\x08\x63hannels\x18\x01 \x03(\x0b\x32\x12.boltlight.Channel\"\xb1\x03\n\x07\x43hannel\x12\x15\n\rremote_pubkey\x18\x01 \x01(\t\x12\x18\n\x10short_channel_id\x18\x02 \x01(\t\x12\x12\n\nchannel_id\x18\x03 \x01(\t\x12\x14\n\x0c\x66unding_txid\x18\x04 \x01(\t\x12\x15\n\rcapacity_msat\x18\x05 \x01(\x04\x12\x1a\n\x12local_balance_msat\x18\x06 \x01(\x04\x12\x1b\n\x13remote_balance_msat\x18\x07 \x01(\x04\x12\x19\n\x11local_reserve_sat\x18\x08 \x01(\x04\x12\x1a\n\x12remote_reserve_sat\x18\t \x01(\x04\x12\x15\n\rto_self_delay\x18\n \x01(\r\x12\x0f\n\x07private\x18\x0b \x01(\x08\x12\'\n\x05state\x18\x0c \x01(\x0e\x32\x18.boltlight.Channel.State\x12\x0e\n\x06\x61\x63tive\x18\r \x01(\x08\"c\n\x05State\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x10\n\x0cPENDING_OPEN\x10\x01\x12\x08\n\x04OPEN\x10\x02\x12\x18\n\x14PENDING_MUTUAL_CLOSE\x10\x03\x12\x17\n\x13PENDING_FORCE_CLOSE\x10\x04\"\xe5\x01\n\x13ListInvoicesRequest\x12\x11\n\tmax_items\x18\x01 \x01(\r\x12\x18\n\x10search_timestamp\x18\x02 \x01(\x04\x12\x30\n\x0csearch_order\x18\x03 \x01(\x0e\x32\x1a.boltlight.Order.Direction\x12.\n\nlist_order\x18\x04 \x01(\x0e\x32\x1a.boltlight.Order.Direction\x12\x0c\n\x04paid\x18\x05 \x01(\x08\x12\x0f\n\x07pending\x18\x06 \x01(\x08\x12\x0f\n\x07\x65xpired\x18\x07 \x01(\x08\x12\x0f\n\x07unknown\x18\x08 \x01(\x08\"<\n\x14ListInvoicesResponse\x12$\n\x08invoices\x18\x01 \x03(\x0b\x32\x12.boltlight.Invoice\"3\n\x05Order\"*\n\tDirection\x12\r\n\tASCENDING\x10\x00\x12\x0e\n\nDESCENDING\x10\x01\"\xea\x02\n\x07Invoice\x12\x1b\n\x13\x61mount_encoded_msat\x18\x01 \x01(\x04\x12\x1c\n\x14\x61mount_received_msat\x18\x02 \x01(\x04\x12\x14\n\x0cpayment_hash\x18\x03 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x04 \x01(\t\x12\x18\n\x10\x64\x65scription_hash\x18\x05 \x01(\t\x12\x11\n\ttimestamp\x18\x06 \x01(\x04\x12\x0e\n\x06\x65xpiry\x18\x07 \x01(\r\x12\x15\n\rfallback_addr\x18\x08 \x01(\t\x12)\n\x0broute_hints\x18\t \x03(\x0b\x32\x14.boltlight.RouteHint\x12\'\n\x05state\x18\n \x01(\x0e\x32\x18.boltlight.Invoice.State\x12\x17\n\x0fpayment_request\x18\x0b \x01(\t\"8\n\x05State\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x08\n\x04PAID\x10\x01\x12\x0b\n\x07PENDING\x10\x02\x12\x0b\n\x07\x45XPIRED\x10\x03\"\x15\n\x13ListPaymentsRequest\"<\n\x14ListPaymentsResponse\x12$\n\x08payments\x18\x01 \x03(\x0b\x32\x12.boltlight.Payment\"s\n\x07Payment\x12\x13\n\x0b\x61mount_msat\x18\x01 \x01(\x04\x12\x11\n\ttimestamp\x18\x02 \x01(\x04\x12\x14\n\x0cpayment_hash\x18\x03 \x01(\t\x12\x18\n\x10payment_preimage\x18\x04 \x01(\t\x12\x10\n\x08\x66\x65\x65_msat\x18\x05 \x01(\x04\"\x12\n\x10ListPeersRequest\"3\n\x11ListPeersResponse\x12\x1e\n\x05peers\x18\x01 \x03(\x0b\x32\x0f.boltlight.Peer\"E\n\x04Peer\x12\x0e\n\x06pubkey\x18\x01 \x01(\t\x12\x0f\n\x07\x61\x64\x64ress\x18\x02 \x01(\t\x12\r\n\x05\x61lias\x18\x03 \x01(\t\x12\r\n\x05\x63olor\x18\x04 \x01(\t\"\x19\n\x17ListTransactionsRequest\"H\n\x18ListTransactionsResponse\x12,\n\x0ctransactions\x18\x01 \x03(\x0b\x32\x16.boltlight.Transaction\"\x94\x01\n\x0bTransaction\x12\x12\n\namount_sat\x18\x01 \x01(\x03\x12\x0c\n\x04txid\x18\x02 \x01(\t\x12\x15\n\rconfirmations\x18\x03 \x01(\r\x12\x14\n\x0c\x62lock_height\x18\x04 \x01(\r\x12\x12\n\nblock_hash\x18\x05 \x01(\t\x12\x11\n\ttimestamp\x18\x06 \x01(\x04\x12\x0f\n\x07\x66\x65\x65_sat\x18\x07 \x01(\x04\"?\n\x11NewAddressRequest\x12*\n\taddr_type\x18\x01 \x01(\x0e\x32\x17.boltlight.Address.Type\"%\n\x12NewAddressResponse\x12\x0f\n\x07\x61\x64\x64ress\x18\x01 \x01(\t\"T\n\x07\x41\x64\x64ress\"I\n\x04Type\x12\n\n\x06P2WPKH\x10\x00\x12\x11\n\rNATIVE_SEGWIT\x10\x00\x12\x0b\n\x07NP2WPKH\x10\x01\x12\x11\n\rNESTED_SEGWIT\x10\x01\x1a\x02\x10\x01\"_\n\x12OpenChannelRequest\x12\x10\n\x08node_uri\x18\x01 \x01(\t\x12\x13\n\x0b\x66unding_sat\x18\x02 \x01(\x04\x12\x11\n\tpush_msat\x18\x03 \x01(\x04\x12\x0f\n\x07private\x18\x04 \x01(\x08\"+\n\x13OpenChannelResponse\x12\x14\n\x0c\x66unding_txid\x18\x01 \x01(\t\"q\n\x11PayInvoiceRequest\x12\x17\n\x0fpayment_request\x18\x01 \x01(\t\x12\x13\n\x0b\x61mount_msat\x18\x02 \x01(\x04\x12\x13\n\x0b\x64\x65scription\x18\x03 \x01(\t\x12\x19\n\x11\x63ltv_expiry_delta\x18\x04 \x01(\r\".\n\x12PayInvoiceResponse\x12\x18\n\x10payment_preimage\x18\x01 \x01(\t\"N\n\x11PayOnChainRequest\x12\x0f\n\x07\x61\x64\x64ress\x18\x01 \x01(\t\x12\x12\n\namount_sat\x18\x02 \x01(\x04\x12\x14\n\x0c\x66\x65\x65_sat_byte\x18\x03 \x01(\r\"\"\n\x12PayOnChainResponse\x12\x0c\n\x04txid\x18\x01 \x01(\t\"%\n\x11UnlockNodeRequest\x12\x10\n\x08password\x18\x01 \x01(\t\"\x14\n\x12UnlockNodeResponse\"\x17\n\x15\x42\x61lanceOnChainRequest\"B\n\x16\x42\x61lanceOnChainResponse\x12\x15\n\rconfirmed_sat\x18\x01 \x01(\x04\x12\x11\n\ttotal_sat\x18\x02 \x01(\x04\x32I\n\x08Unlocker\x12=\n\x06Unlock\x12\x18.boltlight.UnlockRequest\x1a\x19.boltlight.UnlockResponse2\x86\x01\n\tBoltlight\x12@\n\x07GetInfo\x12\x19.boltlight.GetInfoRequest\x1a\x1a.boltlight.GetInfoResponse\x12\x37\n\x04Lock\x12\x16.boltlight.LockRequest\x1a\x17.boltlight.LockResponse2\xe6\n\n\tLightning\x12X\n\x0f\x42\x61lanceOffChain\x12!.boltlight.BalanceOffChainRequest\x1a\".boltlight.BalanceOffChainResponse\x12U\n\x0e\x42\x61lanceOnChain\x12 .boltlight.BalanceOnChainRequest\x1a!.boltlight.BalanceOnChainResponse\x12O\n\x0c\x43heckInvoice\x12\x1e.boltlight.CheckInvoiceRequest\x1a\x1f.boltlight.CheckInvoiceResponse\x12O\n\x0c\x43loseChannel\x12\x1e.boltlight.CloseChannelRequest\x1a\x1f.boltlight.CloseChannelResponse\x12R\n\rCreateInvoice\x12\x1f.boltlight.CreateInvoiceRequest\x1a .boltlight.CreateInvoiceResponse\x12R\n\rDecodeInvoice\x12\x1f.boltlight.DecodeInvoiceRequest\x1a .boltlight.DecodeInvoiceResponse\x12L\n\x0bGetNodeInfo\x12\x1d.boltlight.GetNodeInfoRequest\x1a\x1e.boltlight.GetNodeInfoResponse\x12O\n\x0cListChannels\x12\x1e.boltlight.ListChannelsRequest\x1a\x1f.boltlight.ListChannelsResponse\x12O\n\x0cListInvoices\x12\x1e.boltlight.ListInvoicesRequest\x1a\x1f.boltlight.ListInvoicesResponse\x12O\n\x0cListPayments\x12\x1e.boltlight.ListPaymentsRequest\x1a\x1f.boltlight.ListPaymentsResponse\x12\x46\n\tListPeers\x12\x1b.boltlight.ListPeersRequest\x1a\x1c.boltlight.ListPeersResponse\x12[\n\x10ListTransactions\x12\".boltlight.ListTransactionsRequest\x1a#.boltlight.ListTransactionsResponse\x12I\n\nNewAddress\x12\x1c.boltlight.NewAddressRequest\x1a\x1d.boltlight.NewAddressResponse\x12L\n\x0bOpenChannel\x12\x1d.boltlight.OpenChannelRequest\x1a\x1e.boltlight.OpenChannelResponse\x12I\n\nPayInvoice\x12\x1c.boltlight.PayInvoiceRequest\x1a\x1d.boltlight.PayInvoiceResponse\x12I\n\nPayOnChain\x12\x1c.boltlight.PayOnChainRequest\x1a\x1d.boltlight.PayOnChainResponse\x12I\n\nUnlockNode\x12\x1c.boltlight.UnlockNodeRequest\x1a\x1d.boltlight.UnlockNodeResponseb\x06proto3')



_UNLOCKREQUEST = DESCRIPTOR.message_types_by_name['UnlockRequest']
_UNLOCKRESPONSE = DESCRIPTOR.message_types_by_name['UnlockResponse']
_GETINFOREQUEST = DESCRIPTOR.message_types_by_name['GetInfoRequest']
_GETINFORESPONSE = DESCRIPTOR.message_types_by_name['GetInfoResponse']
_LOCKREQUEST = DESCRIPTOR.message_types_by_name['LockRequest']
_LOCKRESPONSE = DESCRIPTOR.message_types_by_name['LockResponse']
_BALANCEOFFCHAINREQUEST = DESCRIPTOR.message_types_by_name['BalanceOffChainRequest']
_BALANCEOFFCHAINRESPONSE = DESCRIPTOR.message_types_by_name['BalanceOffChainResponse']
_CHECKINVOICEREQUEST = DESCRIPTOR.message_types_by_name['CheckInvoiceRequest']
_CHECKINVOICERESPONSE = DESCRIPTOR.message_types_by_name['CheckInvoiceResponse']
_CLOSECHANNELREQUEST = DESCRIPTOR.message_types_by_name['CloseChannelRequest']
_CLOSECHANNELRESPONSE = DESCRIPTOR.message_types_by_name['CloseChannelResponse']
_CREATEINVOICEREQUEST = DESCRIPTOR.message_types_by_name['CreateInvoiceRequest']
_CREATEINVOICERESPONSE = DESCRIPTOR.message_types_by_name['CreateInvoiceResponse']
_GETNODEINFOREQUEST = DESCRIPTOR.message_types_by_name['GetNodeInfoRequest']
_DECODEINVOICEREQUEST = DESCRIPTOR.message_types_by_name['DecodeInvoiceRequest']
_DECODEINVOICERESPONSE = DESCRIPTOR.message_types_by_name['DecodeInvoiceResponse']
_ROUTEHINT = DESCRIPTOR.message_types_by_name['RouteHint']
_HOPHINT = DESCRIPTOR.message_types_by_name['HopHint']
_NETWORK = DESCRIPTOR.message_types_by_name['Network']
_GETNODEINFORESPONSE = DESCRIPTOR.message_types_by_name['GetNodeInfoResponse']
_LISTCHANNELSREQUEST = DESCRIPTOR.message_types_by_name['ListChannelsRequest']
_LISTCHANNELSRESPONSE = DESCRIPTOR.message_types_by_name['ListChannelsResponse']
_CHANNEL = DESCRIPTOR.message_types_by_name['Channel']
_LISTINVOICESREQUEST = DESCRIPTOR.message_types_by_name['ListInvoicesRequest']
_LISTINVOICESRESPONSE = DESCRIPTOR.message_types_by_name['ListInvoicesResponse']
_ORDER = DESCRIPTOR.message_types_by_name['Order']
_INVOICE = DESCRIPTOR.message_types_by_name['Invoice']
_LISTPAYMENTSREQUEST = DESCRIPTOR.message_types_by_name['ListPaymentsRequest']
_LISTPAYMENTSRESPONSE = DESCRIPTOR.message_types_by_name['ListPaymentsResponse']
_PAYMENT = DESCRIPTOR.message_types_by_name['Payment']
_LISTPEERSREQUEST = DESCRIPTOR.message_types_by_name['ListPeersRequest']
_LISTPEERSRESPONSE = DESCRIPTOR.message_types_by_name['ListPeersResponse']
_PEER = DESCRIPTOR.message_types_by_name['Peer']
_LISTTRANSACTIONSREQUEST = DESCRIPTOR.message_types_by_name['ListTransactionsRequest']
_LISTTRANSACTIONSRESPONSE = DESCRIPTOR.message_types_by_name['ListTransactionsResponse']
_TRANSACTION = DESCRIPTOR.message_types_by_name['Transaction']
_NEWADDRESSREQUEST = DESCRIPTOR.message_types_by_name['NewAddressRequest']
_NEWADDRESSRESPONSE = DESCRIPTOR.message_types_by_name['NewAddressResponse']
_ADDRESS = DESCRIPTOR.message_types_by_name['Address']
_OPENCHANNELREQUEST = DESCRIPTOR.message_types_by_name['OpenChannelRequest']
_OPENCHANNELRESPONSE = DESCRIPTOR.message_types_by_name['OpenChannelResponse']
_PAYINVOICEREQUEST = DESCRIPTOR.message_types_by_name['PayInvoiceRequest']
_PAYINVOICERESPONSE = DESCRIPTOR.message_types_by_name['PayInvoiceResponse']
_PAYONCHAINREQUEST = DESCRIPTOR.message_types_by_name['PayOnChainRequest']
_PAYONCHAINRESPONSE = DESCRIPTOR.message_types_by_name['PayOnChainResponse']
_UNLOCKNODEREQUEST = DESCRIPTOR.message_types_by_name['UnlockNodeRequest']
_UNLOCKNODERESPONSE = DESCRIPTOR.message_types_by_name['UnlockNodeResponse']
_BALANCEONCHAINREQUEST = DESCRIPTOR.message_types_by_name['BalanceOnChainRequest']
_BALANCEONCHAINRESPONSE = DESCRIPTOR.message_types_by_name['BalanceOnChainResponse']
_NETWORK_NAME = _NETWORK.enum_types_by_name['Name']
_CHANNEL_STATE = _CHANNEL.enum_types_by_name['State']
_ORDER_DIRECTION = _ORDER.enum_types_by_name['Direction']
_INVOICE_STATE = _INVOICE.enum_types_by_name['State']
_ADDRESS_TYPE = _ADDRESS.enum_types_by_name['Type']
UnlockRequest = _reflection.GeneratedProtocolMessageType('UnlockRequest', (_message.Message,), {
  'DESCRIPTOR' : _UNLOCKREQUEST,
  '__module__' : 'boltlight.boltlight_pb2'
  # @@protoc_insertion_point(class_scope:boltlight.UnlockRequest)
  })
_sym_db.RegisterMessage(UnlockRequest)

UnlockResponse = _reflection.GeneratedProtocolMessageType('UnlockResponse', (_message.Message,), {
  'DESCRIPTOR' : _UNLOCKRESPONSE,
  '__module__' : 'boltlight.boltlight_pb2'
  # @@protoc_insertion_point(class_scope:boltlight.UnlockResponse)
  })
_sym_db.RegisterMessage(UnlockResponse)

GetInfoRequest = _reflection.GeneratedProtocolMessageType('GetInfoRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETINFOREQUEST,
  '__module__' : 'boltlight.boltlight_pb2'
  # @@protoc_insertion_point(class_scope:boltlight.GetInfoRequest)
  })
_sym_db.RegisterMessage(GetInfoRequest)

GetInfoResponse = _reflection.GeneratedProtocolMessageType('GetInfoResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETINFORESPONSE,
  '__module__' : 'boltlight.boltlight_pb2'
  # @@protoc_insertion_point(class_scope:boltlight.GetInfoResponse)
  })
_sym_db.RegisterMessage(GetInfoResponse)

LockRequest = _reflection.GeneratedProtocolMessageType('LockRequest', (_message.Message,), {
  'DESCRIPTOR' : _LOCKREQUEST,
  '__module__' : 'boltlight.boltlight_pb2'
  # @@protoc_insertion_point(class_scope:boltlight.LockRequest)
  })
_sym_db.RegisterMessage(LockRequest)

LockResponse = _reflection.GeneratedProtocolMessageType('LockResponse', (_message.Message,), {
  'DESCRIPTOR' : _LOCKRESPONSE,
  '__module__' : 'boltlight.boltlight_pb2'
  # @@protoc_insertion_point(class_scope:boltlight.LockResponse)
  })
_sym_db.RegisterMessage(LockResponse)

BalanceOffChainRequest = _reflection.GeneratedProtocolMessageType('BalanceOffChainRequest', (_message.Message,), {
  'DESCRIPTOR' : _BALANCEOFFCHAINREQUEST,
  '__module__' : 'boltlight.boltlight_pb2'
  # @@protoc_insertion_point(class_scope:boltlight.BalanceOffChainRequest)
  })
_sym_db.RegisterMessage(BalanceOffChainRequest)

BalanceOffChainResponse = _reflection.GeneratedProtocolMessageType('BalanceOffChainResponse', (_message.Message,), {
  'DESCRIPTOR' : _BALANCEOFFCHAINRESPONSE,
  '__module__' : 'boltlight.boltlight_pb2'
  # @@protoc_insertion_point(class_scope:boltlight.BalanceOffChainResponse)
  })
_sym_db.RegisterMessage(BalanceOffChainResponse)

CheckInvoiceRequest = _reflection.GeneratedProtocolMessageType('CheckInvoiceRequest', (_message.Message,), {
  'DESCRIPTOR' : _CHECKINVOICEREQUEST,
  '__module__' : 'boltlight.boltlight_pb2'
  # @@protoc_insertion_point(class_scope:boltlight.CheckInvoiceRequest)
  })
_sym_db.RegisterMessage(CheckInvoiceRequest)

CheckInvoiceResponse = _reflection.GeneratedProtocolMessageType('CheckInvoiceResponse', (_message.Message,), {
  'DESCRIPTOR' : _CHECKINVOICERESPONSE,
  '__module__' : 'boltlight.boltlight_pb2'
  # @@protoc_insertion_point(class_scope:boltlight.CheckInvoiceResponse)
  })
_sym_db.RegisterMessage(CheckInvoiceResponse)

CloseChannelRequest = _reflection.GeneratedProtocolMessageType('CloseChannelRequest', (_message.Message,), {
  'DESCRIPTOR' : _CLOSECHANNELREQUEST,
  '__module__' : 'boltlight.boltlight_pb2'
  # @@protoc_insertion_point(class_scope:boltlight.CloseChannelRequest)
  })
_sym_db.RegisterMessage(CloseChannelRequest)

CloseChannelResponse = _reflection.GeneratedProtocolMessageType('CloseChannelResponse', (_message.Message,), {
  'DESCRIPTOR' : _CLOSECHANNELRESPONSE,
  '__module__' : 'boltlight.boltlight_pb2'
  # @@protoc_insertion_point(class_scope:boltlight.CloseChannelResponse)
  })
_sym_db.RegisterMessage(CloseChannelResponse)

CreateInvoiceRequest = _reflection.GeneratedProtocolMessageType('CreateInvoiceRequest', (_message.Message,), {
  'DESCRIPTOR' : _CREATEINVOICEREQUEST,
  '__module__' : 'boltlight.boltlight_pb2'
  # @@protoc_insertion_point(class_scope:boltlight.CreateInvoiceRequest)
  })
_sym_db.RegisterMessage(CreateInvoiceRequest)

CreateInvoiceResponse = _reflection.GeneratedProtocolMessageType('CreateInvoiceResponse', (_message.Message,), {
  'DESCRIPTOR' : _CREATEINVOICERESPONSE,
  '__module__' : 'boltlight.boltlight_pb2'
  # @@protoc_insertion_point(class_scope:boltlight.CreateInvoiceResponse)
  })
_sym_db.RegisterMessage(CreateInvoiceResponse)

GetNodeInfoRequest = _reflection.GeneratedProtocolMessageType('GetNodeInfoRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETNODEINFOREQUEST,
  '__module__' : 'boltlight.boltlight_pb2'
  # @@protoc_insertion_point(class_scope:boltlight.GetNodeInfoRequest)
  })
_sym_db.RegisterMessage(GetNodeInfoRequest)

DecodeInvoiceRequest = _reflection.GeneratedProtocolMessageType('DecodeInvoiceRequest', (_message.Message,), {
  'DESCRIPTOR' : _DECODEINVOICEREQUEST,
  '__module__' : 'boltlight.boltlight_pb2'
  # @@protoc_insertion_point(class_scope:boltlight.DecodeInvoiceRequest)
  })
_sym_db.RegisterMessage(DecodeInvoiceRequest)

DecodeInvoiceResponse = _reflection.GeneratedProtocolMessageType('DecodeInvoiceResponse', (_message.Message,), {
  'DESCRIPTOR' : _DECODEINVOICERESPONSE,
  '__module__' : 'boltlight.boltlight_pb2'
  # @@protoc_insertion_point(class_scope:boltlight.DecodeInvoiceResponse)
  })
_sym_db.RegisterMessage(DecodeInvoiceResponse)

RouteHint = _reflection.GeneratedProtocolMessageType('RouteHint', (_message.Message,), {
  'DESCRIPTOR' : _ROUTEHINT,
  '__module__' : 'boltlight.boltlight_pb2'
  # @@protoc_insertion_point(class_scope:boltlight.RouteHint)
  })
_sym_db.RegisterMessage(RouteHint)

HopHint = _reflection.GeneratedProtocolMessageType('HopHint', (_message.Message,), {
  'DESCRIPTOR' : _HOPHINT,
  '__module__' : 'boltlight.boltlight_pb2'
  # @@protoc_insertion_point(class_scope:boltlight.HopHint)
  })
_sym_db.RegisterMessage(HopHint)

Network = _reflection.GeneratedProtocolMessageType('Network', (_message.Message,), {
  'DESCRIPTOR' : _NETWORK,
  '__module__' : 'boltlight.boltlight_pb2'
  # @@protoc_insertion_point(class_scope:boltlight.Network)
  })
_sym_db.RegisterMessage(Network)

GetNodeInfoResponse = _reflection.GeneratedProtocolMessageType('GetNodeInfoResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETNODEINFORESPONSE,
  '__module__' : 'boltlight.boltlight_pb2'
  # @@protoc_insertion_point(class_scope:boltlight.GetNodeInfoResponse)
  })
_sym_db.RegisterMessage(GetNodeInfoResponse)

ListChannelsRequest = _reflection.GeneratedProtocolMessageType('ListChannelsRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTCHANNELSREQUEST,
  '__module__' : 'boltlight.boltlight_pb2'
  # @@protoc_insertion_point(class_scope:boltlight.ListChannelsRequest)
  })
_sym_db.RegisterMessage(ListChannelsRequest)

ListChannelsResponse = _reflection.GeneratedProtocolMessageType('ListChannelsResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTCHANNELSRESPONSE,
  '__module__' : 'boltlight.boltlight_pb2'
  # @@protoc_insertion_point(class_scope:boltlight.ListChannelsResponse)
  })
_sym_db.RegisterMessage(ListChannelsResponse)

Channel = _reflection.GeneratedProtocolMessageType('Channel', (_message.Message,), {
  'DESCRIPTOR' : _CHANNEL,
  '__module__' : 'boltlight.boltlight_pb2'
  # @@protoc_insertion_point(class_scope:boltlight.Channel)
  })
_sym_db.RegisterMessage(Channel)

ListInvoicesRequest = _reflection.GeneratedProtocolMessageType('ListInvoicesRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTINVOICESREQUEST,
  '__module__' : 'boltlight.boltlight_pb2'
  # @@protoc_insertion_point(class_scope:boltlight.ListInvoicesRequest)
  })
_sym_db.RegisterMessage(ListInvoicesRequest)

ListInvoicesResponse = _reflection.GeneratedProtocolMessageType('ListInvoicesResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTINVOICESRESPONSE,
  '__module__' : 'boltlight.boltlight_pb2'
  # @@protoc_insertion_point(class_scope:boltlight.ListInvoicesResponse)
  })
_sym_db.RegisterMessage(ListInvoicesResponse)

Order = _reflection.GeneratedProtocolMessageType('Order', (_message.Message,), {
  'DESCRIPTOR' : _ORDER,
  '__module__' : 'boltlight.boltlight_pb2'
  # @@protoc_insertion_point(class_scope:boltlight.Order)
  })
_sym_db.RegisterMessage(Order)

Invoice = _reflection.GeneratedProtocolMessageType('Invoice', (_message.Message,), {
  'DESCRIPTOR' : _INVOICE,
  '__module__' : 'boltlight.boltlight_pb2'
  # @@protoc_insertion_point(class_scope:boltlight.Invoice)
  })
_sym_db.RegisterMessage(Invoice)

ListPaymentsRequest = _reflection.GeneratedProtocolMessageType('ListPaymentsRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTPAYMENTSREQUEST,
  '__module__' : 'boltlight.boltlight_pb2'
  # @@protoc_insertion_point(class_scope:boltlight.ListPaymentsRequest)
  })
_sym_db.RegisterMessage(ListPaymentsRequest)

ListPaymentsResponse = _reflection.GeneratedProtocolMessageType('ListPaymentsResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTPAYMENTSRESPONSE,
  '__module__' : 'boltlight.boltlight_pb2'
  # @@protoc_insertion_point(class_scope:boltlight.ListPaymentsResponse)
  })
_sym_db.RegisterMessage(ListPaymentsResponse)

Payment = _reflection.GeneratedProtocolMessageType('Payment', (_message.Message,), {
  'DESCRIPTOR' : _PAYMENT,
  '__module__' : 'boltlight.boltlight_pb2'
  # @@protoc_insertion_point(class_scope:boltlight.Payment)
  })
_sym_db.RegisterMessage(Payment)

ListPeersRequest = _reflection.GeneratedProtocolMessageType('ListPeersRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTPEERSREQUEST,
  '__module__' : 'boltlight.boltlight_pb2'
  # @@protoc_insertion_point(class_scope:boltlight.ListPeersRequest)
  })
_sym_db.RegisterMessage(ListPeersRequest)

ListPeersResponse = _reflection.GeneratedProtocolMessageType('ListPeersResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTPEERSRESPONSE,
  '__module__' : 'boltlight.boltlight_pb2'
  # @@protoc_insertion_point(class_scope:boltlight.ListPeersResponse)
  })
_sym_db.RegisterMessage(ListPeersResponse)

Peer = _reflection.GeneratedProtocolMessageType('Peer', (_message.Message,), {
  'DESCRIPTOR' : _PEER,
  '__module__' : 'boltlight.boltlight_pb2'
  # @@protoc_insertion_point(class_scope:boltlight.Peer)
  })
_sym_db.RegisterMessage(Peer)

ListTransactionsRequest = _reflection.GeneratedProtocolMessageType('ListTransactionsRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTTRANSACTIONSREQUEST,
  '__module__' : 'boltlight.boltlight_pb2'
  # @@protoc_insertion_point(class_scope:boltlight.ListTransactionsRequest)
  })
_sym_db.RegisterMessage(ListTransactionsRequest)

ListTransactionsResponse = _reflection.GeneratedProtocolMessageType('ListTransactionsResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTTRANSACTIONSRESPONSE,
  '__module__' : 'boltlight.boltlight_pb2'
  # @@protoc_insertion_point(class_scope:boltlight.ListTransactionsResponse)
  })
_sym_db.RegisterMessage(ListTransactionsResponse)

Transaction = _reflection.GeneratedProtocolMessageType('Transaction', (_message.Message,), {
  'DESCRIPTOR' : _TRANSACTION,
  '__module__' : 'boltlight.boltlight_pb2'
  # @@protoc_insertion_point(class_scope:boltlight.Transaction)
  })
_sym_db.RegisterMessage(Transaction)

NewAddressRequest = _reflection.GeneratedProtocolMessageType('NewAddressRequest', (_message.Message,), {
  'DESCRIPTOR' : _NEWADDRESSREQUEST,
  '__module__' : 'boltlight.boltlight_pb2'
  # @@protoc_insertion_point(class_scope:boltlight.NewAddressRequest)
  })
_sym_db.RegisterMessage(NewAddressRequest)

NewAddressResponse = _reflection.GeneratedProtocolMessageType('NewAddressResponse', (_message.Message,), {
  'DESCRIPTOR' : _NEWADDRESSRESPONSE,
  '__module__' : 'boltlight.boltlight_pb2'
  # @@protoc_insertion_point(class_scope:boltlight.NewAddressResponse)
  })
_sym_db.RegisterMessage(NewAddressResponse)

Address = _reflection.GeneratedProtocolMessageType('Address', (_message.Message,), {
  'DESCRIPTOR' : _ADDRESS,
  '__module__' : 'boltlight.boltlight_pb2'
  # @@protoc_insertion_point(class_scope:boltlight.Address)
  })
_sym_db.RegisterMessage(Address)

OpenChannelRequest = _reflection.GeneratedProtocolMessageType('OpenChannelRequest', (_message.Message,), {
  'DESCRIPTOR' : _OPENCHANNELREQUEST,
  '__module__' : 'boltlight.boltlight_pb2'
  # @@protoc_insertion_point(class_scope:boltlight.OpenChannelRequest)
  })
_sym_db.RegisterMessage(OpenChannelRequest)

OpenChannelResponse = _reflection.GeneratedProtocolMessageType('OpenChannelResponse', (_message.Message,), {
  'DESCRIPTOR' : _OPENCHANNELRESPONSE,
  '__module__' : 'boltlight.boltlight_pb2'
  # @@protoc_insertion_point(class_scope:boltlight.OpenChannelResponse)
  })
_sym_db.RegisterMessage(OpenChannelResponse)

PayInvoiceRequest = _reflection.GeneratedProtocolMessageType('PayInvoiceRequest', (_message.Message,), {
  'DESCRIPTOR' : _PAYINVOICEREQUEST,
  '__module__' : 'boltlight.boltlight_pb2'
  # @@protoc_insertion_point(class_scope:boltlight.PayInvoiceRequest)
  })
_sym_db.RegisterMessage(PayInvoiceRequest)

PayInvoiceResponse = _reflection.GeneratedProtocolMessageType('PayInvoiceResponse', (_message.Message,), {
  'DESCRIPTOR' : _PAYINVOICERESPONSE,
  '__module__' : 'boltlight.boltlight_pb2'
  # @@protoc_insertion_point(class_scope:boltlight.PayInvoiceResponse)
  })
_sym_db.RegisterMessage(PayInvoiceResponse)

PayOnChainRequest = _reflection.GeneratedProtocolMessageType('PayOnChainRequest', (_message.Message,), {
  'DESCRIPTOR' : _PAYONCHAINREQUEST,
  '__module__' : 'boltlight.boltlight_pb2'
  # @@protoc_insertion_point(class_scope:boltlight.PayOnChainRequest)
  })
_sym_db.RegisterMessage(PayOnChainRequest)

PayOnChainResponse = _reflection.GeneratedProtocolMessageType('PayOnChainResponse', (_message.Message,), {
  'DESCRIPTOR' : _PAYONCHAINRESPONSE,
  '__module__' : 'boltlight.boltlight_pb2'
  # @@protoc_insertion_point(class_scope:boltlight.PayOnChainResponse)
  })
_sym_db.RegisterMessage(PayOnChainResponse)

UnlockNodeRequest = _reflection.GeneratedProtocolMessageType('UnlockNodeRequest', (_message.Message,), {
  'DESCRIPTOR' : _UNLOCKNODEREQUEST,
  '__module__' : 'boltlight.boltlight_pb2'
  # @@protoc_insertion_point(class_scope:boltlight.UnlockNodeRequest)
  })
_sym_db.RegisterMessage(UnlockNodeRequest)

UnlockNodeResponse = _reflection.GeneratedProtocolMessageType('UnlockNodeResponse', (_message.Message,), {
  'DESCRIPTOR' : _UNLOCKNODERESPONSE,
  '__module__' : 'boltlight.boltlight_pb2'
  # @@protoc_insertion_point(class_scope:boltlight.UnlockNodeResponse)
  })
_sym_db.RegisterMessage(UnlockNodeResponse)

BalanceOnChainRequest = _reflection.GeneratedProtocolMessageType('BalanceOnChainRequest', (_message.Message,), {
  'DESCRIPTOR' : _BALANCEONCHAINREQUEST,
  '__module__' : 'boltlight.boltlight_pb2'
  # @@protoc_insertion_point(class_scope:boltlight.BalanceOnChainRequest)
  })
_sym_db.RegisterMessage(BalanceOnChainRequest)

BalanceOnChainResponse = _reflection.GeneratedProtocolMessageType('BalanceOnChainResponse', (_message.Message,), {
  'DESCRIPTOR' : _BALANCEONCHAINRESPONSE,
  '__module__' : 'boltlight.boltlight_pb2'
  # @@protoc_insertion_point(class_scope:boltlight.BalanceOnChainResponse)
  })
_sym_db.RegisterMessage(BalanceOnChainResponse)

_UNLOCKER = DESCRIPTOR.services_by_name['Unlocker']
_BOLTLIGHT = DESCRIPTOR.services_by_name['Boltlight']
_LIGHTNING = DESCRIPTOR.services_by_name['Lightning']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _ADDRESS_TYPE._options = None
  _ADDRESS_TYPE._serialized_options = b'\020\001'
  _UNLOCKREQUEST._serialized_start=40
  _UNLOCKREQUEST._serialized_end=94
  _UNLOCKRESPONSE._serialized_start=96
  _UNLOCKRESPONSE._serialized_end=135
  _GETINFOREQUEST._serialized_start=137
  _GETINFOREQUEST._serialized_end=153
  _GETINFORESPONSE._serialized_start=155
  _GETINFORESPONSE._serialized_end=240
  _LOCKREQUEST._serialized_start=242
  _LOCKREQUEST._serialized_end=255
  _LOCKRESPONSE._serialized_start=257
  _LOCKRESPONSE._serialized_end=271
  _BALANCEOFFCHAINREQUEST._serialized_start=273
  _BALANCEOFFCHAINREQUEST._serialized_end=297
  _BALANCEOFFCHAINRESPONSE._serialized_start=299
  _BALANCEOFFCHAINRESPONSE._serialized_end=418
  _CHECKINVOICEREQUEST._serialized_start=420
  _CHECKINVOICEREQUEST._serialized_end=463
  _CHECKINVOICERESPONSE._serialized_start=465
  _CHECKINVOICERESPONSE._serialized_end=528
  _CLOSECHANNELREQUEST._serialized_start=530
  _CLOSECHANNELREQUEST._serialized_end=586
  _CLOSECHANNELRESPONSE._serialized_start=588
  _CLOSECHANNELRESPONSE._serialized_end=632
  _CREATEINVOICEREQUEST._serialized_start=635
  _CREATEINVOICEREQUEST._serialized_end=769
  _CREATEINVOICERESPONSE._serialized_start=771
  _CREATEINVOICERESPONSE._serialized_end=861
  _GETNODEINFOREQUEST._serialized_start=863
  _GETNODEINFOREQUEST._serialized_end=883
  _DECODEINVOICEREQUEST._serialized_start=885
  _DECODEINVOICEREQUEST._serialized_end=953
  _DECODEINVOICERESPONSE._serialized_start=956
  _DECODEINVOICERESPONSE._serialized_end=1229
  _ROUTEHINT._serialized_start=1231
  _ROUTEHINT._serialized_end=1281
  _HOPHINT._serialized_start=1284
  _HOPHINT._serialized_end=1422
  _NETWORK._serialized_start=1424
  _NETWORK._serialized_end=1493
  _NETWORK_NAME._serialized_start=1435
  _NETWORK_NAME._serialized_end=1493
  _GETNODEINFORESPONSE._serialized_start=1496
  _GETNODEINFORESPONSE._serialized_end=1654
  _LISTCHANNELSREQUEST._serialized_start=1656
  _LISTCHANNELSREQUEST._serialized_end=1698
  _LISTCHANNELSRESPONSE._serialized_start=1700
  _LISTCHANNELSRESPONSE._serialized_end=1760
  _CHANNEL._serialized_start=1763
  _CHANNEL._serialized_end=2196
  _CHANNEL_STATE._serialized_start=2097
  _CHANNEL_STATE._serialized_end=2196
  _LISTINVOICESREQUEST._serialized_start=2199
  _LISTINVOICESREQUEST._serialized_end=2428
  _LISTINVOICESRESPONSE._serialized_start=2430
  _LISTINVOICESRESPONSE._serialized_end=2490
  _ORDER._serialized_start=2492
  _ORDER._serialized_end=2543
  _ORDER_DIRECTION._serialized_start=2501
  _ORDER_DIRECTION._serialized_end=2543
  _INVOICE._serialized_start=2546
  _INVOICE._serialized_end=2908
  _INVOICE_STATE._serialized_start=2852
  _INVOICE_STATE._serialized_end=2908
  _LISTPAYMENTSREQUEST._serialized_start=2910
  _LISTPAYMENTSREQUEST._serialized_end=2931
  _LISTPAYMENTSRESPONSE._serialized_start=2933
  _LISTPAYMENTSRESPONSE._serialized_end=2993
  _PAYMENT._serialized_start=2995
  _PAYMENT._serialized_end=3110
  _LISTPEERSREQUEST._serialized_start=3112
  _LISTPEERSREQUEST._serialized_end=3130
  _LISTPEERSRESPONSE._serialized_start=3132
  _LISTPEERSRESPONSE._serialized_end=3183
  _PEER._serialized_start=3185
  _PEER._serialized_end=3254
  _LISTTRANSACTIONSREQUEST._serialized_start=3256
  _LISTTRANSACTIONSREQUEST._serialized_end=3281
  _LISTTRANSACTIONSRESPONSE._serialized_start=3283
  _LISTTRANSACTIONSRESPONSE._serialized_end=3355
  _TRANSACTION._serialized_start=3358
  _TRANSACTION._serialized_end=3506
  _NEWADDRESSREQUEST._serialized_start=3508
  _NEWADDRESSREQUEST._serialized_end=3571
  _NEWADDRESSRESPONSE._serialized_start=3573
  _NEWADDRESSRESPONSE._serialized_end=3610
  _ADDRESS._serialized_start=3612
  _ADDRESS._serialized_end=3696
  _ADDRESS_TYPE._serialized_start=3623
  _ADDRESS_TYPE._serialized_end=3696
  _OPENCHANNELREQUEST._serialized_start=3698
  _OPENCHANNELREQUEST._serialized_end=3793
  _OPENCHANNELRESPONSE._serialized_start=3795
  _OPENCHANNELRESPONSE._serialized_end=3838
  _PAYINVOICEREQUEST._serialized_start=3840
  _PAYINVOICEREQUEST._serialized_end=3953
  _PAYINVOICERESPONSE._serialized_start=3955
  _PAYINVOICERESPONSE._serialized_end=4001
  _PAYONCHAINREQUEST._serialized_start=4003
  _PAYONCHAINREQUEST._serialized_end=4081
  _PAYONCHAINRESPONSE._serialized_start=4083
  _PAYONCHAINRESPONSE._serialized_end=4117
  _UNLOCKNODEREQUEST._serialized_start=4119
  _UNLOCKNODEREQUEST._serialized_end=4156
  _UNLOCKNODERESPONSE._serialized_start=4158
  _UNLOCKNODERESPONSE._serialized_end=4178
  _BALANCEONCHAINREQUEST._serialized_start=4180
  _BALANCEONCHAINREQUEST._serialized_end=4203
  _BALANCEONCHAINRESPONSE._serialized_start=4205
  _BALANCEONCHAINRESPONSE._serialized_end=4271
  _UNLOCKER._serialized_start=4273
  _UNLOCKER._serialized_end=4346
  _BOLTLIGHT._serialized_start=4349
  _BOLTLIGHT._serialized_end=4483
  _LIGHTNING._serialized_start=4486
  _LIGHTNING._serialized_end=5868
# @@protoc_insertion_point(module_scope)



# Boltlight - a LN node wrapper
#
# Copyright (C) 2021-2022 boltlight contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For a full list of contributors, please see the AUTHORS.md file.
"""Boltlight package."""

import importlib.metadata

__version__ = importlib.metadata.version('boltlight')



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['boltlight',
 'boltlight.migrations',
 'boltlight.migrations.versions',
 'boltlight.utils']

package_data = \
{'': ['*'], 'boltlight': ['share/*']}

install_requires = \
['alembic>=1.5.5,<2.0.0',
 'click>=8.0.4,<9.0.0',
 'googleapis-common-protos>=1.53.0,<2.0.0',
 'grpcio>=1.43.0,<2.0.0',
 'lnd-proto==0.13.3-beta.0',
 'macaroonbakery>=1.3.1,<2.0.0',
 'protobuf>=3.19.1,<4.0.0',
 'pylibscrypt>=1.8.0,<2.0.0',
 'pyln-client==0.10.1',
 'pymacaroons>=0.13.0,<0.14.0',
 'pynacl>=1.3.0,<2.0.0',
 'qrcode>=6.1,<7.0',
 'requests>=2.25.1,<3.0.0',
 'sqlalchemy>=1.3.23,<1.4.0']

entry_points = \
{'console_scripts': ['blink = boltlight.blink:entrypoint',
                     'boltlight = boltlight.boltlight:start',
                     'boltlight-pairing = boltlight.pairing:start',
                     'boltlight-secure = boltlight.secure:secure']}

setup_kwargs = {
    'name': 'boltlight',
    'version': '2.1.0rc1',
    'description': 'Lightning Network node wrapper',
    'long_description': '# Boltlight - a BOLT-on interface to the Lightning Network\n\nBoltlight is a Lightning Network node wrapper.\n\nIt is not a LN node itself and connects to an existing node of one of the\nsupported implementations, providing a uniform interface and set of features.\nClient code that uses boltlight can thus be agnostic on which node is running\nunder the hood.\n\nThis means that the underlying LN node implementation can be\nchanged anytime with minimal intervention and no effects on client code.\n\nEach underlying implementation implements some features with "little"\ndifferences. Boltlight strives to keep a uniform interface at all times,\ndrawing a common line where implementations differ and always choosing to stay\nBOLT-compliant as much as possible.\n\nLAPP developers should be free to code, without the need to lock-in to any\nparticular implementation.\n\n\n### Supported LN implementations :zap:\n\nCurrently, the main LN implementations <sup>1</sup> are supported:\n\n- [c-lightning](https://github.com/ElementsProject/lightning)\n  (v0.10.1) by Blockstream\n- [eclair](https://github.com/ACINQ/eclair) (v0.6.1) by Acinq\n- [electrum](https://github.com/spesmilo/electrum) (v4.1.5)\n  by Thomas Voegtlin\n- [lnd](https://github.com/lightningnetwork/lnd) (v0.13.3-beta) by Lightning\n  Labs\n\n### How it works\n\nOn the client side, boltlight exposes a gRPC client interface, by default on\nport 1708. On the node side, it proxies all the received calls to the\nunderlying implementation, using the appropriate transport and authentication\nto connect and applying the appropriate format and data translations to each\nimplemented call.\n\nSee [Supported APIs](/doc/supported_apis.md) for a table of the supported calls\nfor each implementation.\n\nCalls that are not yet implemented return an `UNIMPLEMENTED` error.\n\nSoftware dependencies and configurations are the only significant differences\nbetween the supported implementations.\n\nSee [Implementation Specific](/doc/implementation_specific.md) for an\nincomplete list of configuration tips and nuances that are dependent on the\nparticular lightning implementation.\n\n\n#### Notes\n1. _at the moment, only the specified versions of the LN nodes are supported_\n\n\n## Requirements\n\nFirst of all, boltlight will need to connect to an existing and supported LN\nnode.\n\nIn order to run boltlight it needs to be configured and software dependencies\nhave to be met. Some dependencies and configuration choices are determined by\nthe implementation of choice. Availability of the required dependencies will be\nchecked at runtime.\n\nThe supported architectures are `amd64` and `arm32v7`. The latter may require\nadditional dependencies.\n\n\n### System dependencies\n\n- Linux <sup>1,2</sup>\n- Python 3.9+ <sup>3</sup>\n- [optional] libscrypt 1.8+ (_faster start_)\n\n### Resources\n\nResource demand should be pretty low:\n\n- CPU: 1 core\n- RAM: ~64MB when idling\n- disk: docker image weights ~350MB\n\n#### Notes\n\n1. _tested on Debian 10 Buster_\n2. _macOS should work but is not well tested, Windows may work but is not\n   actively supported_\n3. _tested with python 3.9+_\n\n\n## Installing\n\nBoltlight can be installed with:\n```\n$ pip install boltlight\n```\n\nUsage of a `virtualenv` is recommended.\n\n\n## Configuring and securing\n\nBoltlight needs to be configured before it can connect to a LN node and make\nitself useful. [Configuring](/doc/configuring.md) contains all the instructions\non how to configure boltlight.\n\nOne last step is necessary after configuration is complete: security.\n\nOn the node side, as with configuration, this partially depends on the chosen\nimplementation. On the client side, authorization is handled with gRPC\nmacaroons.\n\nFor faster setup and execution, TLS and macaroons can be turned off. This is\nnot suitable for production use but can be useful for testing or development.\n\nTo configure the necessary secrets and set a password to manage and protect\nthem run:\n```bash\n$ boltlight-secure\n```\n\nIt can be run interactively or unattended and will create or update boltlight\'s\ndatabase and macaroon files.\n\nAll secrets will be encrypted and stored in the database, so they won\'t be\naccessible at rest. Secrets are decrypted with boltlight\'s password as part\nof the unlocking process, will only be available to boltlight at runtime and\nare never written to disk in plaintext.\n\nPlease read [Security](/doc/security.md) for all the details.\n\n\n## Using\n\n### Running\n\nTo start boltlight run:\n```bash\n$ boltlight\n```\n\nThis will start its gRPC server, through which Boltlight can be operated.\n\nThe `boltlight.proto` defines the structure for the data to be serialized\nand can be found [here](/boltlight/boltlight.proto).\ngRPC client libraries for supported languages can be generated from this file.\nSee the [Generation instructions](/doc/client_libraries.md) for more\ninformation.\n\nThe proto file contains three services: _Unlocker_, _Lightning_ and _Locker_.\nThe first does not require macaroon authorization and can only unlock\nboltlight, the other two start after boltlight is unlocked, require\nauthorization and provide all the supported operations, locking included.\n\n### API documentation\n\nDocumentation for the gRPC APIs along with example code in Python, Go,\nNode.js and Bash can be found at\n[API page](https://hashbeam.gitlab.io/boltlight/).\n\n### CLI\n\nA CLI named `blink`, with bash and zsh completion support, is available for\ninteractive usage, maintenance or testing.\n\nTo activate completion support add the appropriate line to your shell\'s RC\nfile:\n- `~/.bashrc`: `. /path/to/installed/boltlight/share/complete-blink.bash`\n- `~/.zshrc`: `. /path/to/installed/boltlight/share/complete-blink.zsh`\n\nFor a full list of the available CLI commands use the `--help` option:\n```bash\n$ blink --help\n```\n\n### Pairing\n\nTo **pair** boltlight with a client, run:\n```bash\n$ boltlight-pairing\n```\n\nThis will generate two URIs that allow easy retrieval of connection data\n(`boltlightconnect://<host>:<port>[?cert=<PEM certificate>]`)\nand macaroon (`macaroon:<chosen macaroon>`).\nThe URIs can be displayed as QR codes, suitable for easy reading from client\napps, or as text for easy copying.\n\nThe [Globular](https://gitlab.com/hashbeam/globular) Android wallet is an\nexample app that supports pairing with boltlight.\n\n\n## The fork\n\nBoltlight has been forked from [Lighter](https://gitlab.com/inbitcoin/lighter)\nand continues on the same path, aiming to:\n- make it easier to get started\n- add support for more implementations\n- keep the existing support up-to-date\n- add more features\n\nBoltlight is not 100% backwards-compatible with Lighter because of the name\nchange, so a re-configuration is needed in order to upgrade.\n\n\n## Contributing\n\nAll contributions are welcome!\n\nIf you\'re a developer and want to get involved, see\n[CONTRIBUTING.md](/CONTRIBUTING.md) for info on how to join the effort.\n\nIf boltlight is missing that key feature you need, please get in touch.\nFeel free to open issues to request new features or report bugs, or send an\nemail to `hashbeam@protonmail.com`.\n\n',
    'author': 'Hashbeam',
    'author_email': 'hashbeam@protonmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://hashbeam.gitlab.io/boltlight',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9.0,<4.0.0',
}


setup(**setup_kwargs)



# Boltlight - a LN node wrapper
#
# Copyright (C) 2021 boltlight contributors
# Copyright (C) 2018 inbitcoin s.r.l.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For a full list of contributors, please see the AUTHORS.md file.
"""Boltlight's DB migration module."""

from logging import getLogger
from os import path

from alembic import command

from . import settings as sett
from .utils.db import get_alembic_cfg, init_db

LOGGER = getLogger(__name__)


def migrate():
    """Handle DB migration using alembic."""
    if not path.exists(sett.DB_PATH):
        return
    alembic_cfg = get_alembic_cfg(False)
    init_db(alembic_cfg=alembic_cfg)
    # importing this top level would import None
    from .utils.db import ENGINE  # pylint: disable=import-outside-toplevel
    with ENGINE.begin() as connection:
        # pylint: disable=unsupported-assignment-operation
        alembic_cfg.attributes['connection'] = connection
        # pylint: enable=unsupported-assignment-operation
        command.upgrade(alembic_cfg, 'head')



# Boltlight - a LN node wrapper
#
# Copyright (C) 2021 boltlight contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For a full list of contributors, please see the AUTHORS.md file.
"""Utils module for boltlight.py."""

from grpc import StatusCode


class RuntimeTerminate:
    """Provide methods to terminate a runtime gRPC call.

    All methods deny API access to the caller and provide information on the
    reason why it happened.
    """
    @staticmethod
    def missing_macaroon(_ignored_request, context):
        """Terminate gRPC call due to missing authentication."""
        context.abort(StatusCode.UNAUTHENTICATED,
                      'Macaroon authentication missing')

    @staticmethod
    def macaroon_error(_ignored_request, context):
        """Terminate gRPC call due to authentication error."""
        context.abort(StatusCode.PERMISSION_DENIED,
                      'Macaroon authentication error')

    @staticmethod
    def already_unlocked(_ignored_request, context):
        """Terminate unlock gRPC call since boltlight is already unlocked."""
        context.abort(StatusCode.FAILED_PRECONDITION,
                      'Boltlight is already unlocked')

    @staticmethod
    def not_runtime(_ignored_request, context):
        """Terminate gRPC call since the called method does not exist."""
        context.abort(StatusCode.UNIMPLEMENTED, 'Not a runtime method')



# Boltlight - a LN node wrapper
#
# Copyright (C) 2021-2022 boltlight contributors
# Copyright (C) 2018 inbitcoin s.r.l.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Network utils module."""

from importlib import import_module
from logging import getLogger
from time import sleep

from requests import Session as ReqSession
from requests.exceptions import ConnectionError as ReqConnectionErr
from requests.exceptions import Timeout

from .. import boltlight_pb2 as pb
from .. import settings as sett
from ..errors import Err
from .misc import disable_logger

LOGGER = getLogger(__name__)


def check_connection(lock):
    """Check if connection to node is successful by calling GetNodeInfo."""
    try:
        acquired = lock.acquire(blocking=False)
        if not acquired:
            return
        request = pb.GetNodeInfoRequest()
        module = import_module(f'...light_{sett.IMPLEMENTATION}', __name__)
        info = None
        LOGGER.info('Checking connection to %s node...', sett.IMPLEMENTATION)
        attempts = 0
        while not info:
            try:
                with disable_logger():
                    info = getattr(module, 'GetNodeInfo')(request,
                                                          FakeContext())
            except RuntimeError as err:
                LOGGER.error('Connection to LN node failed: %s',
                             str(err).strip())
            attempts += 1
            if not info:
                sleep(min(attempts * 2, 60 * 60))
                continue
            if info.identity_pubkey:
                LOGGER.info('Connection to node "%s" successful',
                            info.identity_pubkey)
            node_ver = module.get_node_version()
            if node_ver:
                LOGGER.info('Using %s version %s', sett.IMPLEMENTATION,
                            node_ver)
            else:
                LOGGER.info('Using %s', sett.IMPLEMENTATION)
    finally:
        if acquired:
            lock.release()


def check_req_params(context, request, *parameters):
    """Raise a missing_parameter error if parameter is missing from request."""
    for param in parameters:
        if not getattr(request, param):
            Err().missing_parameter(context, param)


def get_node_timeout(context, min_time=sett.IMPL_MIN_TIMEOUT):
    """Calculate timeout to use when calling LN node.

    Consider client's gRPC timeout if any.
    """
    node_timeout = min_time
    client_time = context.time_remaining()
    if client_time and client_time > node_timeout:
        node_timeout = client_time - sett.RESPONSE_RESERVED_TIME
    node_timeout = min(sett.IMPL_MAX_TIMEOUT, node_timeout)
    return node_timeout


def get_thread_timeout(context):
    """Calculate timeout for an async thread."""
    wait_time = sett.THREAD_TIMEOUT
    if context.time_remaining():
        # subtracting time to do the request and answer to the client
        wait_time = context.time_remaining() - sett.RESPONSE_RESERVED_TIME
    return max(wait_time, 0)


class FakeContext():  # pylint: disable=too-few-public-methods
    """Simulate a gRPC context.

    This is useful when dealing with methods that require a context but the one
    from the client request isn't available.
    """
    def __init__(self, timeout=None):
        self.timeout = timeout

    @staticmethod
    def abort(scode, msg):
        """Raise a runtime error."""
        assert scode
        raise RuntimeError(msg)

    def time_remaining(self):
        """Return the allowed time remaining for the RPC, in seconds.

        If self.timeout is set to None act as no deadline was specified for the
        RPC.
        """
        return self.timeout


class JSONRPCSession():  # pylint: disable=too-few-public-methods
    """Create and mantain a JSON-RPC session open."""
    def __init__(self, auth=None, headers=None, jsonrpc_ver='2.0'):
        self._session = ReqSession()
        self._auth = auth
        self._headers = {'accept': 'application/json'}
        if isinstance(headers, dict):
            self._headers = {**self._headers, **headers}
        self._jsonrpc_ver = jsonrpc_ver
        self._id_count = 0

    def call(self, context, data=None, url=None, timeout=None):
        """Make an RPC call using the opened session.

        Return the response message and a boolean to signal if the response
        contains an error.
        """
        self._id_count += 1
        if url is None:
            url = sett.RPC_URL
        if timeout is None:
            timeout = get_node_timeout(context)
        tries = sett.RPC_TRIES
        while True:
            try:
                response = self._session.post(url,
                                              data=data,
                                              auth=self._auth,
                                              headers=self._headers,
                                              timeout=(sett.RPC_CONN_TIMEOUT,
                                                       timeout))
                break
            except ReqConnectionErr:
                tries -= 1
                if tries == 0:
                    Err().node_error(context,
                                     'RPC call failed: max retries reached')
                LOGGER.debug(
                    'Connection failed, sleeping for %.1f secs (%d tries '
                    'left)', sett.RPC_SLEEP, tries)
                sleep(sett.RPC_SLEEP)
            except Timeout:
                Err().node_error(context, 'RPC call timed out')
        if response.status_code in (400, 403):
            # when password is wrong return Forbidden, True
            return response.reason, True
        if response.status_code not in (200, 500):
            Err().node_error(
                context,
                f'RPC call failed: {response.status_code} {response.reason}')
        json_response = {}
        try:
            json_response = response.json()
        except Exception:  # pylint: disable=broad-except
            Err().node_error(context, 'RPC response is not encoded in JSON')
        if 'error' in json_response and json_response['error'] is not None:
            err = json_response['error']
            if 'message' in err:
                err = json_response['error']['message']
            LOGGER.debug('RPC err: %s', err)
            return err, True
        if 'result' in json_response:
            LOGGER.debug('RPC res: %s', json_response['result'])
            return json_response['result'], False
        LOGGER.debug('RPC res: %s', json_response)
        return json_response, response.status_code == 500



# Boltlight - a LN node wrapper
#
# Copyright (C) 2021-2022 boltlight contributors
# Copyright (C) 2018 inbitcoin s.r.l.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For a full list of contributors, please see the AUTHORS.md file.
"""Implementation of a CLI to command boltlight.

- Exits with code 0 if everything is OK
- Exits with code 1 when a general client-side error occurs
- Exits with code 64 + <gRPC status code> when a gRPC error is raised by server
(https://github.com/grpc/grpc/blob/master/doc/statuscodes.md)

WARNING: new imports might require updating the package build system.
"""

from codecs import encode
from configparser import Error as ConfigError
from contextlib import contextmanager
from functools import wraps
from json import dumps
from os import getcwd, path

from click import (
    ParamType, argument, echo, group, option, pass_context, version_option)
from click.shell_completion import CompletionItem
from google.protobuf.json_format import MessageToJson
from grpc import (
    FutureTimeoutError, RpcError, channel_ready_future,
    composite_channel_credentials, insecure_channel, metadata_call_credentials,
    secure_channel, ssl_channel_credentials)

from . import __version__
from . import boltlight_pb2 as pb
from . import boltlight_pb2_grpc as pb_grpc
from . import settings as sett

UNLOCKER_APIS = ('Unlock', )
BOLTLIGHT_APIS = (
    'GetInfo',
    'Lock',
)


class AddressType(ParamType):
    """Custom type of Address.Type proto enum."""
    name = 'Address.Type'

    def convert(self, value, param, ctx):
        """Convert input value to corresponding Address.Type enum."""
        if value in ['0', 'NP2WPKH', 'NESTED_SEGWIT']:
            return 0
        if value in ['1', 'P2WPKH', 'NATIVE_SEGWIT']:
            return 1
        return self.fail(f'{value} is not a valid {self.name}', param, ctx)


def _complete_address_type(_ctx, _param, incomplete):
    """Autocomplete AddressType."""
    address_types = {
        'NP2WPKH': 'Pay to nested witness key hash',
        'P2WPKH': 'Pay to witness public key hash'
    }
    return [
        CompletionItem(k, help=v) for k, v in address_types.items()
        if k.startswith(incomplete)
    ]


class OrderDirection(ParamType):
    """Custom type of Order.Direction proto enum."""
    name = 'Order.Direction'

    def convert(self, value, param, ctx):
        """Convert input value to corresponding Order.Direction enum."""
        if value in ['0', 'ASCENDING']:
            return 0
        if value in ['1', 'DESCENDING']:
            return 1
        return self.fail(f'{value} is not a valid {self.name}', param, ctx)


def _complete_order_direction(_ctx, _param, incomplete):
    """Autocomplete OrderDirection."""
    orders = {
        'ASCENDING': 'From oldest to newest',
        'DESCENDING': 'From newest to oldest'
    }
    return [
        CompletionItem(k, help=v) for k, v in orders.items()
        if k.startswith(incomplete)
    ]


def _die(message=None, exit_code=1):
    """Print message to stderr and exit with the requested error code."""
    from .utils.misc import die  # pylint: disable=import-outside-toplevel
    die(message, exit_code)


def _check_rpcserver_addr():
    """Check the rpcserver address, adding port if missing."""
    if not sett.CLI_RPCSERVER:
        _die('Invalid rpcserver address')
    rpcserver = sett.CLI_RPCSERVER.split(':', 1)
    if len(rpcserver) > 1:
        port = rpcserver[1]
        if not port.isdigit():
            _die('Invalid port')
        if int(port) not in range(1, 65536):
            _die('Invalid port')
    else:
        sett.CLI_RPCSERVER = sett.CLI_RPCSERVER + ':' + sett.PORT


def _get_cli_options():
    """Set CLI options."""
    # pylint: disable=import-outside-toplevel
    from .utils.misc import get_config_parser, get_path, set_defaults, str2bool

    # pylint: enable=import-outside-toplevel
    config = get_config_parser(interactive=True)
    c_values = ['RPCSERVER', 'TLSCERT', 'MACAROON', 'INSECURE', 'NO_MACAROON']
    set_defaults(config, c_values)
    sec = 'blink'
    if not sett.CLI_INSECURE:
        sett.CLI_INSECURE = str2bool(config.get(sec, 'INSECURE'))
    if not sett.CLI_NO_MACAROON:
        sett.CLI_NO_MACAROON = str2bool(config.get(sec, 'NO_MACAROON'))
    if not sett.CLI_RPCSERVER:
        sett.CLI_RPCSERVER = config.get(sec, 'RPCSERVER')
        _check_rpcserver_addr()
    if not sett.CLI_INSECURE:
        if not sett.CLI_TLSCERT:
            sett.CLI_TLSCERT = get_path(config.get(sec, 'TLSCERT'))
    else:
        sett.CLI_NO_MACAROON = True
    if not sett.CLI_NO_MACAROON:
        if not sett.CLI_MACAROON:
            sett.CLI_MACAROON = get_path(config.get(sec, 'MACAROON'))


def handle_call(func):
    """Handle a call to boltlight."""

    @wraps(func)
    def wrapper(*args, **kwargs):
        """Get start options and run wrapped function."""
        try:
            _get_cli_options()
            api, req = func(*args, **kwargs)
            stub_name = _get_stub_name(api)
            with _connect(stub_name) as stub:
                res = getattr(stub, api)(req, timeout=sett.CLI_TIMEOUT)
            _print_res(res)
        except RpcError as err:
            # pylint: disable=no-member
            err_code = err.code()
            json_err = {'code': err_code.name, 'details': err.details()}
            # pylint: enable=no-member
            error = dumps(json_err, indent=4, sort_keys=True)
            _die(error, sett.CLI_BASE_GRPC_CODE + err_code.value[0])
        except ConfigError as err:
            _die(f'Configuration error: {err}')

    return wrapper


def _print_res(response):
    """Print response using JSON format."""
    echo(
        MessageToJson(response,
                      including_default_value_fields=True,
                      preserving_proto_field_name=True))


def _get_stub_name(api):
    """Get name of proto service based on API name."""
    if api in UNLOCKER_APIS:
        return 'UnlockerStub'
    if api in BOLTLIGHT_APIS:
        return 'BoltlightStub'
    return 'LightningStub'


@contextmanager
def _connect(stub_class):
    """Connect to boltlight using gRPC (securely or insecurely)."""
    channel = None
    if sett.CLI_INSECURE:
        channel = insecure_channel(sett.CLI_RPCSERVER)
    else:
        if sett.CLI_NO_MACAROON:
            creds = _get_credentials(None)
        else:
            creds = _get_credentials(_metadata_callback)
        channel = secure_channel(sett.CLI_RPCSERVER, creds)
    future_channel = channel_ready_future(channel)
    try:
        future_channel.result(timeout=sett.CLI_TIMEOUT)
    except FutureTimeoutError:
        # Handle gRPC channel that did not connect
        _die('Failed to dial server')
    else:
        stub = getattr(pb_grpc, stub_class)(channel)
        yield stub
        channel.close()


def _get_credentials(callback):
    """Get credentials to open a secure gRPC channel.

    TLS certificate file must exist, macaroons are optional.
    """
    with open(sett.CLI_TLSCERT, 'rb') as file:
        cert = file.read()
    creds = cert_creds = ssl_channel_credentials(root_certificates=cert)
    if callback:  # macaroons are enabled
        if not path.exists(sett.CLI_MACAROON):
            _die('Macaroon file not found')
        auth_creds = metadata_call_credentials(callback)
        creds = composite_channel_credentials(cert_creds, auth_creds)
    return creds


def _metadata_callback(_context, callback):
    """Get boltlight's macaroon to be included in the gRPC request."""
    with open(sett.CLI_MACAROON, 'rb') as file:
        macaroon_bytes = file.read()
        macaroon = encode(macaroon_bytes, 'hex')
    callback([('macaroon', macaroon)], None)


@group()
@option('--config',
        nargs=1,
        help='Path to blink configuration file '
        '(default ~/.boltlight/config).')
@option('--rpcserver',
        nargs=1,
        help='Set host[:port] of boltlight gRPC server.')
@option('--tlscert', nargs=1, help='Path to boltlight\'s TLS certificate.')
@option('--macaroon', nargs=1, help='Path to boltlight\'s macaroon.')
@option('--insecure', is_flag=True, help='Do not use TLS and macaroon.')
@option('--no-macaroon', is_flag=True, help='Do not use macaroon.')
@version_option(version=__version__, message='%(version)s')
@pass_context  # pylint: disable=too-many-arguments
def entrypoint(ctx, config, rpcserver, tlscert, macaroon, insecure,
               no_macaroon):
    """Blink, a CLI for boltlight.

    Paths are relative to the working directory.
    """
    incompatible_opts = {
        'insecure': ['tlscert', 'macaroon'],
        'no_macaroon': ['macaroon'],
    }
    passed_params = [param for param in ctx.params if ctx.params[param]]
    for param in passed_params:
        if param in incompatible_opts:
            if any(opt in passed_params for opt in incompatible_opts[param]):
                _die('Incompatible options')

    if config is not None:
        if not config:
            _die('Invalid configuration file')
        # pylint: disable=import-outside-toplevel
        from .utils.misc import get_path

        # pylint: enable=import-outside-toplevel
        sett.CONFIG = get_path(config, base_path=getcwd())
    if rpcserver is not None:
        sett.CLI_RPCSERVER = rpcserver
        _check_rpcserver_addr()
    if tlscert is not None:
        if not tlscert or not path.exists(tlscert):
            _die(f"Missing TLS certificate '{tlscert}'")
        sett.CLI_TLSCERT = tlscert
    if macaroon is not None:
        if not macaroon or not path.exists(macaroon):
            _die(f"Missing macaroon '{macaroon}'")
        sett.CLI_MACAROON = macaroon
    if insecure:
        sett.CLI_INSECURE = True
    if no_macaroon:
        sett.CLI_NO_MACAROON = True


@entrypoint.command()
@option('--password',
        prompt='Insert boltlight\'s password',
        hide_input=True,
        help='Boltlight\'s password.')
@option('--unlock-node',
        is_flag=True,
        help='Whether to also unlock the LN '
        'node')
@handle_call
def unlock(password, unlock_node):
    """Unlock boltlight."""
    req = pb.UnlockRequest(password=password, unlock_node=unlock_node)
    return 'Unlock', req


@entrypoint.command()
@handle_call
def getinfo():
    """Show info about boltlight and the wrapped implementation."""
    req = pb.GetInfoRequest()
    return 'GetInfo', req


@entrypoint.command()
@handle_call
def lock():
    """Lock boltlight."""
    req = pb.LockRequest()
    return 'Lock', req


@entrypoint.command()
@handle_call
def balanceoffchain():
    """Show the available off-chain balance."""
    req = pb.BalanceOffChainRequest()
    return 'BalanceOffChain', req


@entrypoint.command()
@argument('payment_hash', nargs=1)
@handle_call
def checkinvoice(payment_hash):
    """Check if a LN invoice has been paid."""
    req = pb.CheckInvoiceRequest(payment_hash=payment_hash)
    return 'CheckInvoice', req


@entrypoint.command()
@argument('channel_id')
@option('--force', is_flag=True, help='Whether to force a unilateral close.')
@handle_call
def closechannel(channel_id, force):
    """Close a LN channel."""
    req = pb.CloseChannelRequest(channel_id=channel_id, force=force)
    return 'CloseChannel', req


@entrypoint.command()
@option('--amount-msat', nargs=1, type=int, help='Invoice amount.')
@option('--description', nargs=1, help='Invoice description.')
@option('--expiry',
        nargs=1,
        type=int,
        help='Invoice expiration time, in '
        'seconds (default: 420).')
@option('--min-final-cltv-expiry',
        nargs=1,
        type=int,
        help='CTLV delay '
        '(absolute) to use for the final hop in the route.')
@option('--fallback-addr',
        nargs=1,
        help='Fallback address to use if the LN '
        'payment fails.')
@handle_call
def createinvoice(amount_msat, description, expiry, min_final_cltv_expiry,
                  fallback_addr):
    """Create a LN invoice (BOLT 11)."""
    req = pb.CreateInvoiceRequest(amount_msat=amount_msat,
                                  description=description,
                                  expiry=expiry,
                                  min_final_cltv_expiry=min_final_cltv_expiry,
                                  fallback_addr=fallback_addr)
    return 'CreateInvoice', req


@entrypoint.command()
@argument('payment_request', nargs=1)
@option('--description',
        nargs=1,
        help='Invoice description, whose hash should'
        ' match the description hash in the payment request (if present).')
@handle_call
def decodeinvoice(payment_request, description):
    """Decode a LN invoice (BOLT 11)."""
    req = pb.DecodeInvoiceRequest(payment_request=payment_request,
                                  description=description)
    return 'DecodeInvoice', req


@entrypoint.command()
@handle_call
def getnodeinfo():
    """Show info about the wrapped LN node."""
    req = pb.GetNodeInfoRequest()
    return 'GetNodeInfo', req


@entrypoint.command()
@option('--active-only',
        is_flag=True,
        help='Whether to return active '
        'channels only (channel is open and peer is online).')
@handle_call
def listchannels(active_only):
    """List the node's LN channels."""
    req = pb.ListChannelsRequest(active_only=active_only)
    return 'ListChannels', req


# pylint: disable=too-many-arguments
@entrypoint.command()
@option('--max-items',
        nargs=1,
        type=int,
        help='Maximum number of invoices '
        'to be returned (default: 200).')
@option('--search-timestamp',
        nargs=1,
        type=int,
        help='Timestamp to be used '
        'as starting point for the search.')
@option('--search-order',
        nargs=1,
        type=OrderDirection(),
        shell_complete=_complete_order_direction,
        help='Search direction - requires search_timestamp (default: '
        'ascending).')
@option('--list-order',
        nargs=1,
        type=OrderDirection(),
        shell_complete=_complete_order_direction,
        help='Order of the returned invoices (default: ascending)')
@option('--paid', is_flag=True, help='Whether to include paid invoices.')
@option('--pending', is_flag=True, help='Whether to include pending invoices.')
@option('--expired', is_flag=True, help='Whether to include expired invoices.')
@option('--unknown',
        is_flag=True,
        help='Whether to include invoices with an '
        'unknown state.')
@handle_call  # pylint: enable=too-many-arguments
def listinvoices(max_items, search_timestamp, search_order, list_order, paid,
                 pending, expired, unknown):
    """List the node's LN invoices."""
    req = pb.ListInvoicesRequest(max_items=max_items,
                                 search_timestamp=search_timestamp,
                                 search_order=search_order,
                                 list_order=list_order,
                                 paid=paid,
                                 pending=pending,
                                 expired=expired,
                                 unknown=unknown)
    return 'ListInvoices', req


@entrypoint.command()
@handle_call
def listpayments():
    """List the node's LN payments."""
    req = pb.ListPaymentsRequest()
    return 'ListPayments', req


@entrypoint.command()
@handle_call
def listpeers():
    """List the node's connected peers."""
    req = pb.ListPeersRequest()
    return 'ListPeers', req


@entrypoint.command()
@handle_call
def listtransactions():
    """List the node's on-chain transactions."""
    req = pb.ListTransactionsRequest()
    return 'ListTransactions', req


@entrypoint.command()
@option('--addr-type',
        nargs=1,
        type=AddressType(),
        shell_complete=_complete_address_type,
        help='Bitcoin address type (P2WPKH '
        'or NP2WPKH).')
@handle_call
def newaddress(addr_type):
    """Create a new bitcoin address."""
    req = pb.NewAddressRequest(addr_type=addr_type)
    return 'NewAddress', req


@entrypoint.command()
@argument('node_uri', nargs=1)
@argument('funding_sat', nargs=1, type=int)
@option('--push-msat',
        nargs=1,
        type=int,
        help='Amount (taken from funding_sat'
        ') to be pushed to peer, in millisatoshi.')
@option('--private',
        is_flag=True,
        help='Whether the channel will be private '
        '(not anonunced).')
@handle_call
def openchannel(node_uri, funding_sat, push_msat, private):
    """Connect and open a channel with a peer."""
    req = pb.OpenChannelRequest(node_uri=node_uri,
                                funding_sat=funding_sat,
                                push_msat=push_msat,
                                private=private)
    return 'OpenChannel', req


@entrypoint.command()
@argument('payment_request', nargs=1)
@option('--amount-msat', nargs=1, type=int, help='Value to be paid.')
@option('--description',
        nargs=1,
        help='Invoice description, whose hash should'
        ' match the description hash in the payment request (if present).')
@option('--cltv-expiry-delta',
        nargs=1,
        type=int,
        help='Delta to use for the '
        'time-lock of the CLTV (absolute) extended to the final hop.')
@handle_call
def payinvoice(payment_request, amount_msat, description, cltv_expiry_delta):
    """Pay a LN invoice from its payment request (BOLT 11)."""
    req = pb.PayInvoiceRequest(payment_request=payment_request,
                               amount_msat=amount_msat,
                               description=description,
                               cltv_expiry_delta=cltv_expiry_delta)
    return 'PayInvoice', req


@entrypoint.command()
@argument('address', nargs=1)
@argument('amount_sat', nargs=1, type=int)
@option('--fee-sat-byte',
        nargs=1,
        type=int,
        help='Fee rate in satoshi per '
        'byte.')
@handle_call
def payonchain(address, amount_sat, fee_sat_byte):
    """Pay to a bitcoin address."""
    req = pb.PayOnChainRequest(address=address,
                               amount_sat=amount_sat,
                               fee_sat_byte=fee_sat_byte)
    return 'PayOnChain', req


@entrypoint.command()
@option('--password',
        prompt='Insert boltlight\'s password',
        hide_input=True,
        help='Boltlight\'s password.')
@handle_call
def unlocknode(password):
    """Unlock the wrapped node."""
    req = pb.UnlockNodeRequest(password=password)
    return 'UnlockNode', req


@entrypoint.command()
@handle_call
def balanceonchain():
    """Show the available on-chain balance."""
    req = pb.BalanceOnChainRequest()
    return 'BalanceOnChain', req



# Boltlight - a LN node wrapper
#
# Copyright (C) 2021-2022 boltlight contributors
# Copyright (C) 2018 inbitcoin s.r.l.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For a full list of contributors, please see the AUTHORS.md file.
"""The Python implementation of the gRPC boltlight server."""

import sys
from concurrent.futures import ThreadPoolExecutor
from concurrent.futures import TimeoutError as TimeoutFutError
from configparser import Error as ConfigError
from contextlib import suppress
from functools import wraps
from importlib import import_module
from logging import getLogger
from os import environ
from signal import SIGTERM, signal
from threading import Lock, Thread, active_count
from time import sleep, strftime, time

from grpc import (
    ServerInterceptor, StatusCode, server, ssl_server_credentials,
    unary_unary_rpc_method_handler)
from sqlalchemy.exc import SQLAlchemyError

from . import __version__
from . import boltlight_pb2 as pb
from . import boltlight_pb2_grpc as pb_grpc
from . import settings as sett
from .errors import Err
from .macaroons import check_macaroons, get_baker
from .utils.boltlight import RuntimeTerminate
from .utils.db import (
    detect_impl_secret, get_mac_params_from_db, init_db, is_db_ok,
    session_scope)
from .utils.exceptions import InterruptException
from .utils.misc import (
    die, handle_importerror, handle_keyboardinterrupt, handle_sigterm,
    init_common)
from .utils.network import FakeContext, check_connection, check_req_params
from .utils.security import Crypter, ScryptParams, check_password, get_secret

signal(SIGTERM, handle_sigterm)

LOGGER = getLogger(__name__)

environ["GRPC_SSL_CIPHER_SUITES"] = (
    "HIGH+ECDSA:"
    "ECDHE-RSA-AES128-GCM-SHA256:ECDHE-RSA-AES256-GCM-SHA384")


def _handle_logs(func):
    """Wrap API methods to log received request and returned response."""

    @wraps(func)
    def wrapper(*args, **kwargs):
        """Log information about RPC, before and after its execution."""
        start_time = time()
        peer = user_agent = 'unknown'
        req = args[0]
        ctx = args[1]
        if len(args) == 3:
            req = args[1]
            ctx = args[2]
        with suppress(ValueError):
            peer = ctx.peer().split(':', 1)[1]
        for data in ctx.invocation_metadata():
            if data.key == 'user-agent':
                user_agent = data.value
                break
        LOGGER.info('< %-24s %s %s', req.DESCRIPTOR.name, peer, user_agent)
        res = func(*args, **kwargs)
        res_name = res.DESCRIPTOR.name
        stop_time = time()
        call_time = round(stop_time - start_time, 3)
        LOGGER.info('> %-24s %s %2.3fs', res_name, peer, call_time)
        LOGGER.debug('Full response: %s', str(res).replace('\n', ' '))
        return res

    return wrapper


class UnlockerServicer(pb_grpc.UnlockerServicer):
    """Implementation of the Unlocker service defined by protobuf.

    This service does not require macaroons authentication.
    """

    # pylint: disable=too-few-public-methods

    @_handle_logs
    def Unlock(self, request, context):
        """Unlock boltlight.

        If password is correct, unlock boltlight database and stop the
        UnlockerServicer. Otherwise reject request.
        """
        check_req_params(context, request, 'password')
        mod = import_module(f'..light_{sett.IMPLEMENTATION}', __name__)
        plain_secret = None
        with session_scope(context) as session:
            check_password(context, session, request.password)
            if not sett.DISABLE_MACAROONS:
                mac_params = ScryptParams('')
                mac_params.deserialize(get_mac_params_from_db(session))
                sett.MAC_ROOT_KEY = Crypter.gen_derived_key(
                    request.password, mac_params)
                baker = get_baker(sett.MAC_ROOT_KEY, put_ops=True)
                sett.RUNTIME_BAKER = baker
            if sett.IMPLEMENTATION_SECRETS:
                plain_secret = get_secret(context,
                                          session,
                                          request.password,
                                          sett.IMPLEMENTATION,
                                          sett.IMPL_SEC_TYPE,
                                          active_only=True)
        # Calls the implementation specific update method
        mod.update_settings(plain_secret)
        res = pb.UnlockResponse()
        if request.unlock_node:
            with ThreadPoolExecutor(max_workers=1) as executor:
                try:
                    future = executor.submit(mod.unlock_node, FakeContext(),
                                             request.password)
                    future.result(timeout=1)  # max 1 second to unlock node
                except TimeoutFutError:
                    executor.shutdown(wait=False)
                except RuntimeError as err:
                    # don't fail boltlight unlock if node unlock fails
                    LOGGER.info(err)
                else:
                    res.node_unlocked = True
        sett.UNLOCKER_STOP = True
        return res


class BoltlightServicer(pb_grpc.BoltlightServicer):
    """Implementation of the Boltlight service defined by protobuf.

    This service requires macaroons authentication (if not deactivated) and
    is active only after a successful unlock.
    """

    # pylint: disable=too-few-public-methods

    @_handle_logs
    def GetInfo(self, _req, _ctx):
        """Return info about boltlight and the wrapped implementation."""
        mod = import_module(f'..light_{sett.IMPLEMENTATION}', __name__)
        return pb.GetInfoResponse(version=__version__,
                                  node_implementation=sett.IMPLEMENTATION,
                                  node_version=mod.get_node_version())

    @_handle_logs
    def Lock(self, request, context):
        """
        Lock boltlight by deleting secrets from memory and stopping the
        runtime server (LightningServicer + BoltlightServicer).
        """
        sett.MAC_ROOT_KEY = None
        sett.RUNTIME_BAKER = None
        sett.ECL_PASS = None
        sett.LND_MAC = None
        sett.RUNTIME_STOP = True
        return pb.LockResponse()


class LightningServicer():  # pylint: disable=too-few-public-methods
    """Implementation of the Lightning service defined by protobuf.

    This service requires macaroons authentication (if not deactivated) and
    is active only after a successful unlock.

    Allow dynamic dispatching by not deriving from the protobuf generated
    class.
    """

    def __getattr__(self, name):
        """Dispatch the gRPC request dynamically."""

        @_handle_logs
        def dispatcher(req, ctx):
            """Call the requested API method for the configured implementation.

            If an AttributeError error is raised, terminate the RPC with
            a unimplemented_method error.
            """
            # Importing module for specific implementation
            module = import_module(f'..light_{sett.IMPLEMENTATION}', __name__)
            # Searching client requested function in module
            try:
                func = getattr(module, name)
            except AttributeError:
                Err().unimplemented_method(ctx, name)
            # Return requested function if implemented
            return func(req, ctx)

        return dispatcher


def _runtime_terminator(method):
    """Return an RpcMethodHandler if request is not accepted."""
    return unary_unary_rpc_method_handler(method)


def _service_locked_terminator():
    """Return an RpcMethodHandler if service is locked."""

    def terminate(_ignored_request, ctx):
        """Terminates gRPC call."""
        LOGGER.error('- Not an Unlocker operation')
        ctx.abort(StatusCode.UNAVAILABLE, 'Service is locked')

    return unary_unary_rpc_method_handler(terminate)


def _check_request(handler):
    """Check if request is authorized.

    A request is authorized if it's defined in settings.ALL_PERMS and macaroons
    are valid (unless they are disabled).
    Return None if authorization is granted, error handler method otherwise.
    """
    if handler.method not in sett.ALL_PERMS:
        if handler.method == '/boltlight.Unlocker/Unlock':
            LOGGER.error('- Unlock is not a runtime operation')
            return RuntimeTerminate.already_unlocked
        LOGGER.error('- Not a runtime operation')
        return RuntimeTerminate.not_runtime
    if sett.DISABLE_MACAROONS:
        return None
    return check_macaroons(handler.invocation_metadata, handler.method)


class RuntimeInterceptor(ServerInterceptor):
    """gRPC interceptor for the runtime (Boltlight + Lightning) services."""

    # pylint: disable=too-few-public-methods, no-init

    def intercept_service(self, continuation, handler_call_details):
        """Intercept the RPC to decide if request is authorized."""
        error_method = _check_request(handler_call_details)
        if error_method:
            return _runtime_terminator(error_method)
        return continuation(handler_call_details)


class UnlockerInterceptor(ServerInterceptor):
    """gRPC interceptor for the Unlocker service."""

    # pylint: disable=too-few-public-methods

    def __init__(self):  # pylint: disable=super-init-not-called
        self._terminator = _service_locked_terminator()

    def intercept_service(self, continuation, handler_call_details):
        """Intercept the RPC to eventually inform that service is locked."""
        if handler_call_details.method == \
                '/boltlight.Unlocker/Unlock':
            return continuation(handler_call_details)
        return self._terminator


def _create_server(interceptors):
    """Create a gRPC server attaching given interceptors to server.

    Depending on boltlight configuration, server can be created in secure or
    insecure mode (no TLS nor macaroons).
    """
    if sett.INSECURE_CONNECTION:
        grpc_server = server(ThreadPoolExecutor(max_workers=sett.GRPC_WORKERS),
                             interceptors=interceptors)
        grpc_server.add_insecure_port(sett.LISTEN_ADDR)
    else:
        grpc_server = server(ThreadPoolExecutor(max_workers=sett.GRPC_WORKERS),
                             interceptors=interceptors)
        with open(sett.SERVER_KEY, 'rb') as key:
            private_key = key.read()
        with open(sett.SERVER_CRT, 'rb') as cert:
            certificate_chain = cert.read()
        server_credentials = ssl_server_credentials(((
            private_key,
            certificate_chain,
        ), ))
        grpc_server.add_secure_port(sett.LISTEN_ADDR, server_credentials)
    return grpc_server


def _serve_unlocker():
    """Start the unlocker gRPC server.

    The server must be composed by the Unlocker service.
    """
    grpc_server = _create_server([UnlockerInterceptor()])
    pb_grpc.add_UnlockerServicer_to_server(UnlockerServicer(), grpc_server)
    grpc_server.start()
    _log_listening('Unlocker service')
    LOGGER.info('Waiting for password to unlock Lightning service...')
    _unlocker_wait(grpc_server)


def _serve_runtime():
    """Start the runtime gRPC server.

    The server must be composed by the Boltlight and Lightning services.
    """
    grpc_server = _create_server([RuntimeInterceptor()])
    pb_grpc.add_LightningServicer_to_server(LightningServicer(), grpc_server)
    pb_grpc.add_BoltlightServicer_to_server(BoltlightServicer(), grpc_server)
    sett.RUNTIME_SERVER = grpc_server
    grpc_server.start()
    _log_listening('Lightning service')
    _runtime_wait(grpc_server)


def _log_listening(servicer_name):
    """Log at which host and port the service is listening."""
    if sett.INSECURE_CONNECTION:
        LOGGER.info('%s listening on %s (insecure connection)', servicer_name,
                    sett.LISTEN_ADDR)
    else:
        LOGGER.info('%s listening on %s (secure connection)', servicer_name,
                    sett.LISTEN_ADDR)


def _interrupt_threads():
    """Try to gracefully stop all pending threads of the runtime server."""
    close_event = None
    if sett.RUNTIME_SERVER:
        close_event = sett.RUNTIME_SERVER.stop(sett.GRPC_GRACE_TIME)
    if close_event:
        while not close_event.is_set() or sett.THREADS:
            LOGGER.info('Waiting for %s threads to complete...',
                        active_count())
            sleep(3)
    LOGGER.info('All threads shutdown correctly')


def _unlocker_wait(grpc_server):
    """Wait a signal to stop the unlocker server."""
    while not sett.UNLOCKER_STOP:
        sleep(1)
    grpc_server.stop(0)
    sett.UNLOCKER_STOP = False


def _runtime_wait(grpc_server):
    """Wait a signal to stop the runtime server."""
    while not sett.RUNTIME_STOP:
        sleep(1)
    grpc_server.stop(0)
    sett.RUNTIME_STOP = False


def _log_intro():
    """Print a booting boilerplate to ease run distinction."""
    LOGGER.info(' ' * 72)
    LOGGER.info(' ' * 72)
    LOGGER.info(' ' * 72)
    LOGGER.info('*' * 72)
    LOGGER.info(' ' * 72)
    LOGGER.info('Boltlight')
    LOGGER.info('version %s', __version__)
    LOGGER.info(' ' * 72)
    LOGGER.info('booting up at %s', strftime(sett.LOG_TIMEFMT))
    LOGGER.info(' ' * 72)
    LOGGER.info('*' * 72)


def _log_outro():
    """Print a quitting boilerplate to ease run distinction."""
    LOGGER.info('stopping at %s', strftime(sett.LOG_TIMEFMT))
    LOGGER.info('*' * 37)


def _start_services(lock):
    """Handle the unlocker and the runtime servers start."""
    _serve_unlocker()
    con_thread = Thread(target=check_connection, args=(lock, ))
    con_thread.daemon = True
    con_thread.start()
    _serve_runtime()


@handle_keyboardinterrupt
def _start_boltlight():
    """Start boltlight.

    Check if a module for the requested implementation exists and import it.
    Initialize boltlight and start the Unlocker gRPC service.
    """
    init_common("Start boltlight's gRPC server", runtime=True)
    _log_intro()
    init_db()
    with session_scope(FakeContext()) as session:
        if not is_db_ok(session):
            raise RuntimeError(
                'Your database configuration is incomplete or old. '
                'Update it by running boltlight-secure (and deleting db)')
        sett.IMPLEMENTATION_SECRETS = detect_impl_secret(session)
    lock = Lock()
    while True:
        _start_services(lock)


def start():
    """Boltlight entrypoint.

    Any raised and uncaught exception will be handled here.
    """
    try:
        _start_boltlight()
    except ImportError as err:
        handle_importerror(err)
    except KeyError as err:
        LOGGER.error("The environment variable '%s' needs to be set", err)
        die()
    except RuntimeError as err:
        if str(err):
            LOGGER.error(str(err))
        die()
    except FileNotFoundError as err:
        LOGGER.error(str(err))
        die()
    except ConfigError as err:
        err_msg = ''
        if str(err):
            err_msg = str(err)
        LOGGER.error('Configuration error: %s', err_msg)
        die()
    except SQLAlchemyError as err:
        err_msg = ''
        if str(err):
            err_msg = str(err)
        LOGGER.error('DB error: %s', err_msg)
        die()
    except InterruptException:
        _interrupt_threads()
        _log_outro()
        sys.exit(0)
